from __future__ import annotations
from typing import Type, Union, List, TYPE_CHECKING

from _balder.fixture_execution_level import FixtureExecutionLevel
from _balder.testresult import ResultState, BranchBodyResult
from _balder.executor.basic_executable_executor import BasicExecutableExecutor
from _balder.executor.variation_executor import VariationExecutor
from _balder.previous_executor_mark import PreviousExecutorMark
from _balder.controllers.scenario_controller import ScenarioController

if TYPE_CHECKING:
    from _balder.scenario import Scenario
    from _balder.executor.setup_executor import SetupExecutor
    from _balder.fixture_manager import FixtureManager


class ScenarioExecutor(BasicExecutableExecutor):
    """
    A ScenarioExecutor can contain :meth:`VariationExecutor` as children.
    """
    fixture_execution_level = FixtureExecutionLevel.SCENARIO

    def __init__(self, scenario: Type[Scenario], parent: SetupExecutor):
        super().__init__()
        self._variation_executors: List[VariationExecutor] = []
        # check if instance already exists
        if hasattr(scenario, "_instance") and scenario._instance is not None and \
                isinstance(scenario._instance, scenario):
            self._base_scenario_class = scenario._instance
        else:
            self._base_scenario_class = scenario()
            scenario._instance = self._base_scenario_class
        self._parent_executor = parent
        self._fixture_manager = parent.fixture_manager

        # contains the result object for the BODY part of this branch
        self.body_result = BranchBodyResult(self)

        # if the related scenario class of this executor is decorated with ``@covered_by``, this property contains a
        # list with the :class:`ScenarioExecutor` or/and :class:`TestcaseExecutor` that covers this one
        self.covered_by_executors = None

    # ---------------------------------- STATIC METHODS ----------------------------------------------------------------

    # ---------------------------------- CLASS METHODS ----------------------------------------------------------------

    # ---------------------------------- PROPERTIES --------------------------------------------------------------------

    @property
    def all_child_executors(self) -> List[VariationExecutor]:
        return self._variation_executors

    @property
    def parent_executor(self) -> SetupExecutor:
        return self._parent_executor

    @property
    def base_instance(self) -> object:
        """
        returns the base class instance to which this executor instance belongs
        """
        return self.base_scenario_class

    @property
    def base_scenario_class(self) -> Scenario:
        """returns the :class:`Scenario` class that belongs to this executor"""
        return self._base_scenario_class

    @property
    def base_scenario_controller(self) -> ScenarioController:
        """returns the :class:`ScenarioController` for the setup object of this executor"""
        return ScenarioController.get_for(self.base_scenario_class.__class__)

    @property
    def fixture_manager(self) -> FixtureManager:
        """returns the current active fixture manager that belongs to this scenario executor"""
        return self._fixture_manager

    @property
    def all_run_tests(self) -> List[callable]:
        """returns a list of all test methods that are declared to `RUN` in their base :class:`Scenario` class"""
        return self.base_scenario_controller.get_run_test_methods()

    @property
    def all_skip_tests(self) -> List[callable]:
        """returns a list of all test methods that are declared to `SKIP` in their base :class:`Scenario` class"""
        return self.base_scenario_controller.get_skip_test_methods()

    @property
    def all_ignore_tests(self) -> List[callable]:
        """returns a list of all test methods that are declared to `IGNORE` in their base :class:`Scenario` class"""
        return self.base_scenario_controller.get_ignore_test_methods()

    # ---------------------------------- PROTECTED METHODS -------------------------------------------------------------

    def _prepare_execution(self, show_discarded):
        print(f"  SCENARIO {self.base_scenario_class.__class__.__name__}")

    def _body_execution(self, show_discarded):
        for cur_variation_executor in self.get_variation_executors(return_discarded=show_discarded):
            prev_mark = cur_variation_executor.prev_mark
            if cur_variation_executor.has_runnable_tests(consider_discarded_too=show_discarded) \
                    or cur_variation_executor.has_skipped_tests():
                cur_variation_executor.execute(show_discarded=show_discarded)
            elif prev_mark == PreviousExecutorMark.SKIP:
                cur_variation_executor.set_result_for_whole_branch(ResultState.SKIP)
            elif prev_mark == PreviousExecutorMark.COVERED_BY:
                cur_variation_executor.set_result_for_whole_branch(ResultState.COVERED_BY)
            else:
                cur_variation_executor.set_result_for_whole_branch(ResultState.NOT_RUN)

    def _cleanup_execution(self, show_discarded):
        pass

    # ---------------------------------- METHODS -----------------------------------------------------------------------

    def get_variation_executors(self, return_discarded=False) -> List[VariationExecutor]:
        """
        :param return_discarded: True if the method should return discarded variations too

        :return: returns all variation executors that are child executor of this scenario executor
        """
        if not return_discarded:
            return [cur_executor for cur_executor in self._variation_executors
                    if cur_executor.prev_mark != PreviousExecutorMark.DISCARDED]
        return self._variation_executors

    def cleanup_empty_executor_branches(self, consider_discarded=False):
        """
        This method removes all sub executors that are empty and not relevant anymore.
        """
        to_remove_executor = []
        for cur_variation_executor in self.get_variation_executors(return_discarded=consider_discarded):
            if len(cur_variation_executor.get_testcase_executors()) == 0:
                # remove this whole executor because it has no children anymore
                to_remove_executor.append(cur_variation_executor)
        for cur_variation_executor in to_remove_executor:
            self._variation_executors.remove(cur_variation_executor)

    def get_covered_by_element(self) -> List[Union[Scenario, callable]]:
        """
        This method returns a list of elements where the whole scenario is covered from. This means, that the whole
        test methods in this scenario are already be covered from one of the elements in the list.
        """
        return self.base_scenario_controller.get_abs_covered_by_dict().get(None, [])

    def add_variation_executor(self, variation_executor: VariationExecutor):
        """
        This method adds a new VariationExecutor to the child element list of the tree
        """
        if not isinstance(variation_executor, VariationExecutor):
            raise TypeError("the given object `variation_executor` must be of type `VariationExecutor`")
        if variation_executor in self._variation_executors:
            raise ValueError("the given object `variation_executor` already exists in child list")
        self._variation_executors.append(variation_executor)

    def get_executor_for_device_mapping(self, device_mapping: dict) -> Union[VariationExecutor, None]:
        """
        This method searches for a VariationExecutor in the internal list for which the given device mapping is
        contained in

        :param device_mapping: the device_mapping dictionary for which the executor should be searched for

        :return: returns the associated VariationExecutor or None if no matching could be found
        """
        for cur_variation_executor in self._variation_executors:
            if cur_variation_executor.base_device_mapping == device_mapping:
                return cur_variation_executor
        # can not find some
        return None



from multiprocessing import Process

from _balder.testresult import ResultState
from _balder.balder_session import BalderSession


def test_1_no_scenario_devices(balder_working_dir):
    """
    This testcase uses the CALCULATOR environment in a modified version. For this test, the scenario does not define any
    devices.

    The test secures that the setup will still be executed.
    """
    proc = Process(target=processed, args=(balder_working_dir, ))
    proc.start()
    proc.join()
    assert proc.exitcode == 0, "the process terminates with an error"


def processed(env_dir):
    print("\n", flush=True)
    session = BalderSession(cmd_args=[], working_dir=env_dir)
    session.run()
    print("\n")

    assert session.executor_tree.executor_result == ResultState.SUCCESS, \
        "test session does not terminates with success"
    assert len(session.executor_tree.all_child_executors) == 1, "found not exactly one matching setup"
    setup_executor = session.executor_tree.get_setup_executors()[0]
    assert setup_executor.base_setup_class.__class__.__name__ == "SetupPythonAdd", "unexpected matching setup"
    assert len(setup_executor.get_scenario_executors()) == 1, "found not exactly one matching scenario"
    scenario_executor = setup_executor.get_scenario_executors()[0]
    assert scenario_executor.base_scenario_class.__class__.__name__ == "ScenarioAdding", "unexpected matching scenario"
    assert len(scenario_executor.get_variation_executors()) == 1, "found not exactly one matching scenario"



import balder
import logging
from .setup_features import SetupFeatureI, SetupFeatureII
from ..lib.connections import AConnection
from .setup_a import SetupA

logger = logging.getLogger(__name__)


class SetupAChild(SetupA):
    """This is the CHILD setup of category A (exactly the same as scenario A)"""

    class SetupDevice1(SetupA.SetupDevice1):
        s_i = SetupFeatureI()

    @balder.connect(SetupDevice1, over_connection=AConnection)
    class SetupDevice2(SetupA.SetupDevice2):
        s_ii = SetupFeatureII()



import balder
from ..lib.features import UnusedFeatureForUnusedVDevice
from ..lib.connections import MySimplySharedMemoryConnection
from ..lib.utils import SharedObj
from .features_setup import PyAddCalculate, PyAddProvideANumber


class SetupPythonAdd(balder.Setup):

    class Calculator(balder.Device):
        calc = PyAddCalculate()

    class DoNothingDevice(balder.Device):
        unused = UnusedFeatureForUnusedVDevice()

    @balder.connect(Calculator, over_connection=MySimplySharedMemoryConnection)
    @balder.connect(DoNothingDevice, over_connection=balder.Connection())
    class NumberProvider1(balder.Device):
        n = PyAddProvideANumber(CalculatorDevice="Calculator")

    @balder.connect(Calculator, over_connection=MySimplySharedMemoryConnection)
    class NumberProvider2(balder.Device):
        n = PyAddProvideANumber(CalculatorDevice="Calculator")

    @balder.fixture(level="testcase")
    def cleanup_memory(self):
        SharedObj.shared_mem_list = []



from .feature_i import SetupFeatureI
from .feature_i_2 import SetupFeatureI2
from .feature_ii import SetupFeatureII
from .feature_ii_2 import SetupFeatureII2

__all__ = [
    "SetupFeatureI",
    "SetupFeatureI2",
    "SetupFeatureII",
    "SetupFeatureII2",
]



import balder
from balder.parametrization import Parameter
from ..lib.features import AddCalculateFeature, ProvidesANumberFeature, RandomNumberFeature


class ScenarioAdding(balder.Scenario):

    class Calculator(balder.Device):
        # we could add vDevices here later too -> f.e. `NumberOne="NumberOneDevice", NumberTwo="NumberTwoDevice"`
        adds = AddCalculateFeature()

    @balder.connect(Calculator, over_connection=balder.Connection)
    class NumberOneDevice(balder.Device):
        number = ProvidesANumberFeature()
        example = RandomNumberFeature()

    @balder.connect(Calculator, over_connection=balder.Connection)
    class NumberTwoDevice(balder.Device):
        number = ProvidesANumberFeature()
        example = RandomNumberFeature()

    @balder.parametrize_by_feature('number1', (NumberOneDevice, 'example', 'get_random_numbers_higher_than'), parameter={'value': Parameter('number2')})
    @balder.parametrize_by_feature('number2', (NumberTwoDevice, 'example', 'get_random_numbers_higher_than'), parameter={'value': Parameter('number1')})
    def test_add_two_numbers(self, number1, number2):
        self.NumberOneDevice.number.set_number(number1)
        self.NumberTwoDevice.number.set_number(number2)

        self.NumberOneDevice.number.sends_the_number()
        self.NumberTwoDevice.number.sends_the_number()

        self.Calculator.adds.get_numbers()
        result = self.Calculator.adds.add_numbers()
        assert result == 7, "result is not as expected"



import balder
import logging
from ..lib.features import FeatureVDeviceI, FeatureVDeviceII, FeatureVDeviceIII
from ..lib.connections import AConnection, BConnection, CConnection
from .setup_features import SetupMethVarFeature
from ..balderglob import RuntimeObserver

logger = logging.getLogger(__name__)


class SetupIIIAOrB(balder.Setup):
    """This is a setup that uses VDeviceIII and only has a connection `AConnection OR B`"""

    class SetupDevice1(balder.Device):
        s_i = FeatureVDeviceI()
        s_ii = FeatureVDeviceII()
        s_iii = FeatureVDeviceIII()

    @balder.connect(SetupDevice1, over_connection=balder.Connection.based_on(AConnection | BConnection))
    class SetupDevice2(balder.Device):
        s_ii = SetupMethVarFeature(VDeviceIII="SetupDevice1")

    @balder.fixture(level="session")
    def fixture_session(self):
        RuntimeObserver.add_entry(__file__, SetupIIIAOrB, SetupIIIAOrB.fixture_session, category="fixture",
                                  part="construction", msg="begin execution CONSTRUCTION of fixture")

        self.SetupDevice1.s_i.do_something()
        self.SetupDevice2.s_ii.do_something()

        yield

        RuntimeObserver.add_entry(__file__, SetupIIIAOrB, SetupIIIAOrB.fixture_session, category="fixture",
                                  part="teardown", msg="begin execution TEARDOWN of fixture")

        self.SetupDevice1.s_i.do_something()
        self.SetupDevice2.s_ii.do_something()

    @balder.fixture(level="setup")
    def fixture_setup(self):
        RuntimeObserver.add_entry(__file__, SetupIIIAOrB, SetupIIIAOrB.fixture_setup, category="fixture",
                                  part="construction", msg="begin execution CONSTRUCTION of fixture")

        self.SetupDevice1.s_i.do_something()
        self.SetupDevice2.s_ii.do_something()

        yield

        RuntimeObserver.add_entry(__file__, SetupIIIAOrB, SetupIIIAOrB.fixture_setup, category="fixture",
                                  part="teardown", msg="begin execution TEARDOWN of fixture")

        self.SetupDevice1.s_i.do_something()
        self.SetupDevice2.s_ii.do_something()

    @balder.fixture(level="scenario")
    def fixture_scenario(self):
        RuntimeObserver.add_entry(__file__, SetupIIIAOrB, SetupIIIAOrB.fixture_scenario, category="fixture",
                                  part="construction", msg="begin execution CONSTRUCTION of fixture")

        self.SetupDevice1.s_i.do_something()
        self.SetupDevice2.s_ii.do_something()

        yield

        RuntimeObserver.add_entry(__file__, SetupIIIAOrB, SetupIIIAOrB.fixture_scenario, category="fixture",
                                  part="teardown", msg="begin execution TEARDOWN of fixture")

        self.SetupDevice1.s_i.do_something()
        self.SetupDevice2.s_ii.do_something()

    @balder.fixture(level="variation")
    def fixture_variation(self):
        RuntimeObserver.add_entry(__file__, SetupIIIAOrB, SetupIIIAOrB.fixture_variation, category="fixture",
                                  part="construction", msg="begin execution CONSTRUCTION of fixture")

        self.SetupDevice1.s_i.do_something()
        self.SetupDevice2.s_ii.do_something()
        self.SetupDevice2.s_ii.do_something_as_var()

        yield

        RuntimeObserver.add_entry(__file__, SetupIIIAOrB, SetupIIIAOrB.fixture_variation, category="fixture",
                                  part="teardown", msg="begin execution TEARDOWN of fixture")

        self.SetupDevice1.s_i.do_something()
        self.SetupDevice2.s_ii.do_something()
        self.SetupDevice2.s_ii.do_something_as_var()

    @balder.fixture(level="testcase")
    def fixture_testcase(self):
        RuntimeObserver.add_entry(__file__, SetupIIIAOrB, SetupIIIAOrB.fixture_testcase, category="fixture",
                                  part="construction", msg="begin execution CONSTRUCTION of fixture")

        self.SetupDevice1.s_i.do_something()
        self.SetupDevice2.s_ii.do_something()
        self.SetupDevice2.s_ii.do_something_as_var()

        yield

        RuntimeObserver.add_entry(__file__, SetupIIIAOrB, SetupIIIAOrB.fixture_testcase, category="fixture",
                                  part="teardown", msg="begin execution TEARDOWN of fixture")

        self.SetupDevice1.s_i.do_something()
        self.SetupDevice2.s_ii.do_something()
        self.SetupDevice2.s_ii.do_something_as_var()



import balder
from balder import parametrization
import logging
from ..lib.features import FeatureI, FeatureII
from ..lib.connections import AConnection
from ..balderglob import RuntimeObserver

logger = logging.getLogger(__file__)


class ScenarioA(balder.Scenario):
    """This is the scenario of category A"""

    class ScenarioDevice1(balder.Device):
        i = FeatureI()

    @balder.connect(ScenarioDevice1, over_connection=AConnection)
    class ScenarioDevice2(balder.Device):
        ii = FeatureII()

    @balder.parametrize_by_feature(
        'test', (ScenarioDevice2, 'ii', 'get_dynamic_parameters'),
        parameter=dict(
            param1=parametrization.FeatureAccessSelector(ScenarioDevice1, 'i', 'get_1'),
            param2=parametrization.Value(3)
        )
    )
    def test_a_1(self, test):
        RuntimeObserver.add_entry(__file__, ScenarioA, ScenarioA.test_a_1, category="testcase",
                                  msg=f"execute Test `{ScenarioA.test_a_1.__qualname__}` with Parameter `test={test}`")
        self.ScenarioDevice1.i.do_something()
        self.ScenarioDevice2.ii.do_something()

    def test_a_2(self):
        RuntimeObserver.add_entry(__file__, ScenarioA, ScenarioA.test_a_2, category="testcase",
                                  msg=f"execute Test `{ScenarioA.test_a_2.__qualname__}`")
        self.ScenarioDevice1.i.do_something()
        self.ScenarioDevice2.ii.do_something()

    @balder.fixture(level="session")
    def fixture_session(self):
        RuntimeObserver.add_entry(__file__, ScenarioA, ScenarioA.fixture_session, category="fixture",
                                  part="construction", msg="begin execution CONSTRUCTION of fixture")

        self.ScenarioDevice1.i.do_something()
        self.ScenarioDevice2.ii.do_something()

        yield

        RuntimeObserver.add_entry(__file__, ScenarioA, ScenarioA.fixture_session, category="fixture", part="teardown",
                                  msg="begin execution TEARDOWN of fixture")

        self.ScenarioDevice1.i.do_something()
        self.ScenarioDevice2.ii.do_something()

    @balder.fixture(level="setup")
    def fixture_setup(self):
        RuntimeObserver.add_entry(__file__, ScenarioA, ScenarioA.fixture_setup, category="fixture", part="construction",
                                  msg="begin execution CONSTRUCTION of fixture")
        self.ScenarioDevice1.i.do_something()
        self.ScenarioDevice2.ii.do_something()

        yield

        RuntimeObserver.add_entry(__file__, ScenarioA, ScenarioA.fixture_setup, category="fixture", part="teardown",
                                  msg="begin execution TEARDOWN of fixture")

        self.ScenarioDevice1.i.do_something()
        self.ScenarioDevice2.ii.do_something()

    @balder.fixture(level="scenario")
    def fixture_scenario(self):
        RuntimeObserver.add_entry(__file__, ScenarioA, ScenarioA.fixture_scenario, category="fixture",
                                  part="construction", msg="begin execution CONSTRUCTION of fixture")

        self.ScenarioDevice1.i.do_something()
        self.ScenarioDevice2.ii.do_something()

        yield

        RuntimeObserver.add_entry(__file__, ScenarioA, ScenarioA.fixture_scenario, category="fixture", part="teardown",
                                  msg="begin execution TEARDOWN of fixture")

        self.ScenarioDevice1.i.do_something()
        self.ScenarioDevice2.ii.do_something()

    @balder.fixture(level="variation")
    def fixture_variation(self):
        RuntimeObserver.add_entry(__file__, ScenarioA, ScenarioA.fixture_variation, category="fixture",
                                  part="construction", msg="begin execution CONSTRUCTION of fixture")
        self.ScenarioDevice1.i.do_something()
        self.ScenarioDevice2.ii.do_something()

        yield

        RuntimeObserver.add_entry(__file__, ScenarioA, ScenarioA.fixture_variation, category="fixture", part="teardown",
                                  msg="begin execution TEARDOWN of fixture")

        self.ScenarioDevice1.i.do_something()
        self.ScenarioDevice2.ii.do_something()

    @balder.fixture(level="testcase")
    def fixture_testcase(self):
        RuntimeObserver.add_entry(__file__, ScenarioA, ScenarioA.fixture_testcase, category="fixture",
                                  part="construction", msg="begin execution CONSTRUCTION of fixture")

        self.ScenarioDevice1.i.do_something()
        self.ScenarioDevice2.ii.do_something()

        yield

        RuntimeObserver.add_entry(__file__, ScenarioA, ScenarioA.fixture_testcase, category="fixture", part="teardown",
                                  msg="begin execution TEARDOWN of fixture")

        self.ScenarioDevice1.i.do_something()
        self.ScenarioDevice2.ii.do_something()



from __future__ import annotations
from typing import Dict, Tuple, Type

import inspect
from _balder.collector import Collector
from _balder.device import Device
from _balder.parametrization import FeatureAccessSelector, Parameter, Value


def parametrize_by_feature(
        field_name: str,
        feature_accessor: Tuple[Type[Device], str, str],
        parameter: Dict[str, FeatureAccessSelector | Parameter | Value] = None
):
    """
    Allows to parametrize a test function. This decorator will be used to dynamically parametrize a test function, by
    the value a setup feature returns before entering the test.

    :param field_name: the field name of the test function
    :param feature_accessor: a tuple that provides information for accessing the feature
    :param parameter: the parameter to parametrize the feature method (if necessary)
    """
    if not isinstance(field_name, str):
        raise ValueError('the given field name must be a string')
    if parameter is None:
        parameter = {}

    feature_accessor = FeatureAccessSelector(*feature_accessor, parameters=parameter)

    def decorator(func):
        if not inspect.isfunction(func):
            raise TypeError('the decorated object needs to be a test method')

        Collector.register_possible_parametrization(func, field_name, feature_accessor)
        return func
    return decorator



from _balder.controllers.controller import Controller

from _balder.controllers.base_device_controller import BaseDeviceController
from _balder.controllers.device_controller import DeviceController

from _balder.controllers.vdevice_controller import VDeviceController

from _balder.controllers.normal_scenario_setup_controller import NormalScenarioSetupController
from _balder.controllers.scenario_controller import ScenarioController
from _balder.controllers.setup_controller import SetupController

from _balder.controllers.feature_controller import FeatureController



from __future__ import annotations
from typing import List, Type, Union, TYPE_CHECKING

import inspect
from _balder.scenario import Scenario
from _balder.exceptions import InheritanceError

if TYPE_CHECKING:
    from _balder.connection import Connection
    from _balder.cnnrelations.base_connection_relation import BaseConnectionRelationT
    from _balder.utils.typings import MethodLiteralType


def get_scenario_inheritance_list_of(scenario: Type[Scenario]) -> List[Type[Scenario]]:
    """
    This method returns a list, beginning with the given `scenario`, of all parent classes that are from type
    :class:´Scenario`. The last element of the list is always the original :class:`Scenario` class.

    :param scenario: the scenario the method should look for in the inheritance list

    :returns: a list with all classes that are from type scenario
    """
    if scenario == Scenario:
        return [Scenario]
    base_class_of_interest = None
    for cur_parent in scenario.__bases__:
        if issubclass(cur_parent, Scenario):
            if base_class_of_interest is not None:
                raise InheritanceError("it is not allowed to have multiple parent class that inherit from the "
                                       "class `Scenario`")
            base_class_of_interest = cur_parent

    return [scenario] + get_scenario_inheritance_list_of(base_class_of_interest)


def get_class_that_defines_method(meth):
    """credits to https://stackoverflow.com/a/25959545"""
    if inspect.ismethod(meth):
        for cls in inspect.getmro(meth.__self__.__class__):
            if meth.__name__ in cls.__dict__:
                return cls
        meth = meth.__func__  # fallback to __qualname__ parsing
    if inspect.isfunction(meth):
        cls = getattr(inspect.getmodule(meth), meth.__qualname__.split('.<locals>', 1)[0].rsplit('.', 1)[0], None)
        if isinstance(cls, type):
            return cls
    return None  # not required since None would have been implicitly returned anyway


def cnn_type_check_and_convert(elem: Union[Connection, Type[Connection], BaseConnectionRelationT]) \
        -> Union[Connection, BaseConnectionRelationT]:
    """
    converts possible type object to instance and checks if the element is a connection type

    :param elem: the connection object to be converted/checked
    """

    from ..connection import Connection  # pylint: disable=import-outside-toplevel
    from ..cnnrelations.and_connection_relation import AndConnectionRelation  # pylint: disable=import-outside-toplevel
    from ..cnnrelations.or_connection_relation import OrConnectionRelation  # pylint: disable=import-outside-toplevel

    if isinstance(elem, type):
        if issubclass(elem, Connection):
            return elem()
    elif isinstance(elem, (Connection, AndConnectionRelation, OrConnectionRelation)):
        # okay
        return elem
    raise TypeError(f'object needs to be a `Connection`, a connection relation or a `Type[Connection]` - no '
                    f'`{elem}`')


def get_method_type(func_class, func) -> MethodLiteralType:
    """
    This helper function returns the type of the method (`staticmethod`, `classmethod` or `instancemethod`). It never
    returns `function` because this type does not have a class.
    """
    expected_class_qualname = func.__qualname__.rpartition('.')[0]

    def get_for(the_class):
        if func.__name__ in the_class.__dict__.keys() and the_class.__qualname__ == expected_class_qualname:
            fn_type = the_class.__dict__.get(func.__name__)

            if isinstance(fn_type, classmethod):
                return 'classmethod'

            if isinstance(fn_type, staticmethod):
                return 'staticmethod'

            if fn_type is not None and fn_type.__class__.__name__ == 'function':
                return 'instancemethod'
            raise TypeError(f'unknown element type `{func}`')

        for cur_base in the_class.__bases__:
            result = get_for(cur_base)
            if result:
                return result
        return None

    meth_type = get_for(func_class)
    if meth_type is None:
        raise KeyError(f'the provided function `{func.__qualname__}` does not match with the provided class '
                       f'`{func_class.__qualname__}`')
    return meth_type



from multiprocessing import Process

from _balder.testresult import ResultState
from _balder.balder_session import BalderSession


def test_1_setup_inheritance_overwrite_parent_connections(balder_working_dir):
    """
    This testcase executes the general CALCULATOR example.

    It has two inherited setups, while the parent setup defines one completely wrong connection and the child the
    correct matching connection. The environment defines one scenario, that requires the connections, that are provided
    in the child setup.

    The test validates, that only the setup child will match with the scenario and only one variation is
    executed.
    """
    proc = Process(target=processed, args=(balder_working_dir, ))
    proc.start()
    proc.join()
    assert proc.exitcode == 0, "the process terminates with an error"


def processed(env_dir):
    print("\n", flush=True)
    session = BalderSession(cmd_args=[], working_dir=env_dir)
    session.run()
    print("\n")

    assert session.executor_tree.executor_result == ResultState.SUCCESS, \
        "test session does not terminates with success"
    assert len(session.executor_tree.get_setup_executors()) == 1, \
        "not exactly one setup executor found"
    assert session.executor_tree.get_setup_executors()[0].base_setup_class.__class__.__name__ == \
           "SetupPythonAddChild", "wrong scenario class was executed"

    scenario_executors = session.executor_tree.get_setup_executors()[0].get_scenario_executors()

    assert len(scenario_executors) == 1, \
        "not exactly one scenario executor found"
    assert scenario_executors[0].base_scenario_class.__class__.__name__ == "ScenarioAdding", \
        "wrong scenario class was executed"
    assert len(scenario_executors[0].get_variation_executors()) == 1, "not exactly one variation executor found"



import re

from _balder.testresult import ResultState
from _balder.balder_session import BalderSession
from ...test_utilities.base_0_envtester_class import Base0EnvtesterClass


class Test0ResolveOnlyAndShowDiscarded(Base0EnvtesterClass):
    """
    This testcase executes the basic ENV example with a specific command line argument
    ``--resolve-only --show-discarded``.

    The test checks that no tests are executed and the printed information is like expected. Additionally, it ensures
    that all discarded tests are printed too and the reason why they were discarded is printed too.
    """

    @property
    def cmd_args(self):
        return ['--resolve-only', '--show-discarded']

    @property
    def expected_data(self) -> tuple:
        return tuple()

    def validate_printed_output(self, stdout: str) -> bool:
        assert self._check_header_of_stdout(stdout, 2, 2, 2, expected_discarded=6), \
            f"problems within header output"
        remaining_lines = stdout.splitlines()
        assert remaining_lines[7] == "RESOLVING OVERVIEW"
        split_blocks = "\n".join(remaining_lines[8:]).strip().split('\n\n')
        assert len(split_blocks) == 8

        expected_blocks = [
"""XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
X [DISCARDED]  Scenario `ScenarioA` <-> Setup `SetupB`
X    ScenarioA.ScenarioDevice1 = SetupB.SetupDevice3
X    ScenarioA.ScenarioDevice2 = SetupB.SetupDevice4
X    -> Testcase<ScenarioA.test_a_1>
X    -> Testcase<ScenarioA.test_a_2>
X
X    DISCARDED BECAUSE `this variation can not be applicable because there was no setup feature implementation of `FeatureI` (used by scenario device `ScenarioDevice1`) in setup device `SetupDevice3``
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX""",
"""XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
X [DISCARDED]  Scenario `ScenarioA` <-> Setup `SetupB`
X    ScenarioA.ScenarioDevice1 = SetupB.SetupDevice4
X    ScenarioA.ScenarioDevice2 = SetupB.SetupDevice3
X    -> Testcase<ScenarioA.test_a_1>
X    -> Testcase<ScenarioA.test_a_2>
X
X    DISCARDED BECAUSE `this variation can not be applicable because there was no setup feature implementation of `FeatureI` (used by scenario device `ScenarioDevice1`) in setup device `SetupDevice4``
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX""",
"""++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
+ [APPLICABLE] Scenario `ScenarioB` <-> Setup `SetupB`
+    ScenarioB.ScenarioDevice3 = SetupB.SetupDevice3
+    ScenarioB.ScenarioDevice4 = SetupB.SetupDevice4
+    -> Testcase<ScenarioB.test_b_1>
+    -> Testcase<ScenarioB.test_b_2>
++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++""",
"""XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
X [DISCARDED]  Scenario `ScenarioB` <-> Setup `SetupB`
X    ScenarioB.ScenarioDevice3 = SetupB.SetupDevice4
X    ScenarioB.ScenarioDevice4 = SetupB.SetupDevice3
X    -> Testcase<ScenarioB.test_b_1>
X    -> Testcase<ScenarioB.test_b_2>
X
X    DISCARDED BECAUSE `this variation can not be applicable because there was no setup feature implementation of `FeatureIII` (used by scenario device `ScenarioDevice3`) in setup device `SetupDevice4``
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX""",
"""++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
+ [APPLICABLE] Scenario `ScenarioA` <-> Setup `SetupA`
+    ScenarioA.ScenarioDevice1 = SetupA.SetupDevice1
+    ScenarioA.ScenarioDevice2 = SetupA.SetupDevice2
+    -> Testcase<ScenarioA.test_a_1>
+    -> Testcase<ScenarioA.test_a_2>
++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++""",
"""XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
X [DISCARDED]  Scenario `ScenarioA` <-> Setup `SetupA`
X    ScenarioA.ScenarioDevice1 = SetupA.SetupDevice2
X    ScenarioA.ScenarioDevice2 = SetupA.SetupDevice1
X    -> Testcase<ScenarioA.test_a_1>
X    -> Testcase<ScenarioA.test_a_2>
X
X    DISCARDED BECAUSE `this variation can not be applicable because there was no setup feature implementation of `FeatureI` (used by scenario device `ScenarioDevice1`) in setup device `SetupDevice2``
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX""",
"""XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
X [DISCARDED]  Scenario `ScenarioB` <-> Setup `SetupA`
X    ScenarioB.ScenarioDevice3 = SetupA.SetupDevice1
X    ScenarioB.ScenarioDevice4 = SetupA.SetupDevice2
X    -> Testcase<ScenarioB.test_b_1>
X    -> Testcase<ScenarioB.test_b_2>
X
X    DISCARDED BECAUSE `this variation can not be applicable because there was no setup feature implementation of `FeatureIII` (used by scenario device `ScenarioDevice3`) in setup device `SetupDevice1``
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX""",
"""XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
X [DISCARDED]  Scenario `ScenarioB` <-> Setup `SetupA`
X    ScenarioB.ScenarioDevice3 = SetupA.SetupDevice2
X    ScenarioB.ScenarioDevice4 = SetupA.SetupDevice1
X    -> Testcase<ScenarioB.test_b_1>
X    -> Testcase<ScenarioB.test_b_2>
X
X    DISCARDED BECAUSE `this variation can not be applicable because there was no setup feature implementation of `FeatureIII` (used by scenario device `ScenarioDevice3`) in setup device `SetupDevice2``
XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX"""
        ]

        assert sorted(split_blocks) == sorted(expected_blocks), f"detect issues in blocks"
        return True

    @staticmethod
    def validate_finished_session(session: BalderSession):
        assert session.executor_tree.executor_result == ResultState.NOT_RUN, \
            "test session does not terminates with NOT_RUN"



from multiprocessing import Process

from _balder.testresult import ResultState
from _balder.balder_session import BalderSession


def test_1_illegal_non_testmethod_decorated(balder_working_dir):
    """
    This testcase executes the basic example but decorated a non-testmethod with the `covered_by` decorator. The test
    expects an error message.
    """
    proc = Process(target=processed, args=(balder_working_dir, ))
    proc.start()
    proc.join()
    assert proc.exitcode == 0, "the process terminates with an error"


def check_error_of_exc(exc: BaseException):
    assert isinstance(exc, TypeError)
    assert exc.args[0] == ("the use of the `@covered_by` decorator is only allowed for test methods of "
                                     "`Scenario` objects - the method `ScenarioAdding.normal_method` does not "
                                     "start with `test_` and is not a valid test method")

def processed(env_dir):
    print("\n", flush=True)
    session = BalderSession(cmd_args=[], working_dir=env_dir)
    try:
        session.run()
        print("\n")
        assert False, "test session terminates without an error"
    except RuntimeError as exc:
        assert exc.args[0] == ("Error calling __set_name__ on 'CoveredByDecorator' instance 'normal_method' in "
                               "'ScenarioAdding'")
        check_error_of_exc(exc.__cause__)
    except TypeError as exc:
        check_error_of_exc(exc)

    assert session.executor_tree is None, "test session does not terminates before collector work was done"



from __future__ import annotations
from typing import List
from enum import Enum


class FixtureExecutionLevel(Enum):
    """
    This enum describes the fixture-execution-level of a fixture. It describes when the fixture should be executed. This
    level will be set in the fixture decorator.
    """
    SESSION = 'session'
    SETUP = 'setup'
    SCENARIO = 'scenario'
    VARIATION = 'variation'
    TESTCASE = 'testcase'

    @classmethod
    def get_order(cls) -> List[FixtureExecutionLevel]:
        """
        returns the execution order of fixtures
        """
        return [cls.SESSION, cls.SETUP, cls.SCENARIO, cls.VARIATION, cls.TESTCASE]



import balder
import logging
from ..lib.features import FeatureI, FeatureII
from ..lib.connections import BetweenConnection

logger = logging.getLogger(__file__)


class ScenarioA(balder.Scenario):
    """This is the scenario of category A"""

    class ScenarioDevice1(balder.Device):
        i = FeatureI()

    @balder.connect(ScenarioDevice1, over_connection=BetweenConnection)
    class ScenarioDevice2(balder.Device):
        ii = FeatureII()

    def test_a_1(self):
        logger.info(f"execute testcase `{ScenarioA.test_a_1.__qualname__}`")
        assert True

    def test_a_2(self):
        logger.info(f"execute testcase `{ScenarioA.test_a_2.__qualname__}`")
        assert True



from balder.exceptions import DeviceOverwritingError, BalderException
from _balder.balder_session import BalderSession

from tests.test_utilities.base_0_envtester_class import Base0EnvtesterClass


class Test0SetupInheritanceMissingDeviceInheritance(Base0EnvtesterClass):
    """
    This testcase executes a reduced version of the basic envtester environment. It only implements the `ScenarioA` and
    its related `SetupA` and a child class of the related `SetupA`.

    The child class define both devices, but inherits only from one device correctly.

    .. note::
        The `SetupAParent` class has a child (the `SetupAChild` class). This forbids the execution of the
        `SetupAParent` class.
    """

    @property
    def expected_data(self) -> tuple:
        return ()

    @property
    def expected_exit_code(self) -> int:
        return 4

    @staticmethod
    def handle_balder_exception(exc: BalderException):
        assert isinstance(exc, DeviceOverwritingError), 'unexpected error type'
        assert exc.args[0] == "the inner device class `SetupAChild.SetupDevice2` has the same name than the device " \
                              "`SetupAParent.SetupDevice2` - it should also inherit from it"

    @staticmethod
    def validate_finished_session(session: BalderSession):
        assert session.executor_tree is None, "test session does not terminates before collector work was done"



import balder
from ..lib.features import SendMessageFeature, ReceiveMessageFeature


class ScenarioSenderAndReceiver(balder.Scenario):
    class SenderDevice(balder.Device):
        send = SendMessageFeature(Receiver="ReceiverDevice")

    @balder.connect(with_device=SenderDevice, over_connection=balder.Connection())
    class ReceiverDevice(balder.Device):
        recv = ReceiveMessageFeature()

    def test_send_a_message(self):
        msg_content = "my nice message"

        self.ReceiverDevice.recv.start_listening()
        self.SenderDevice.send.send(msg_content)

        assert self.ReceiverDevice.recv.get_next_received_msg() == msg_content, \
            "receiver does not receive the sent message "



import balder
from ...lib.features import FeatureI
from ...balderglob import RuntimeObserver


class SetupFeatureI(FeatureI):

    class OtherVDevice2(balder.VDevice):
        pass

    def do_something(self):
        RuntimeObserver.add_entry(
            __file__, SetupFeatureI, SetupFeatureI.do_something, "enter `SetupFeatureI.do_something`",
            category="feature")



from __future__ import annotations
from typing import TYPE_CHECKING
from abc import ABC, abstractmethod

if TYPE_CHECKING:
    from _balder.scenario import Scenario
    from _balder.executor.scenario_executor import ScenarioExecutor


class MixinCanBeCoveredByExecutor(ABC):
    """mixin class for executor that can hold covered-by items"""

    @property
    @abstractmethod
    def scenario_executor(self) -> ScenarioExecutor:
        """
        returns the current active scenario executor
        """

    @abstractmethod
    def get_covered_by_element(self) -> list[Scenario | callable]:
        """
        This method returns a list of elements where the elements of the executor are covered from.
        """



import balder
import logging
from .setup_features import (SetupFeatureI, SetupFeatureII, SetupFeatureIII, SetupFeatureIV, SetupFeatureV,
                             SetupFeatureVI, SetupFeatureVII, SetupFeatureVIII)
from ..lib.connections import AConnection
from ..balderglob import RuntimeObserver

logger = logging.getLogger(__name__)


class SetupA(balder.Setup):
    """This is a setup of category A (exactly the same as scenario A)"""

    class SetupDevice1(balder.Device):
        s_i = SetupFeatureI()
        s_ii = SetupFeatureII()

    @balder.connect(SetupDevice1, over_connection=AConnection)
    class SetupDevice2(balder.Device):
        s_iii = SetupFeatureIII()
        s_iv = SetupFeatureIV()

    class SetupUnmappedDevice1(balder.Device):
        s_v = SetupFeatureV()
        s_vi = SetupFeatureVI()

    class SetupUnmappedDevice2(balder.Device):
        s_vii = SetupFeatureVII()
        s_viii = SetupFeatureVIII()

    @balder.fixture(level="session")
    def fixture_session(self):
        RuntimeObserver.add_entry(__file__, SetupA, SetupA.fixture_session, category="fixture", part="construction",
                                  msg="begin execution CONSTRUCTION of fixture")

        self.SetupDevice1.s_i.do_something()
        self.SetupDevice1.s_ii.do_something()
        self.SetupDevice2.s_iii.do_something()
        self.SetupDevice2.s_iv.do_something()
        self.SetupUnmappedDevice1.s_v.do_something()
        self.SetupUnmappedDevice1.s_vi.do_something()
        self.SetupUnmappedDevice2.s_vii.do_something()
        self.SetupUnmappedDevice2.s_viii.do_something()

        yield

        RuntimeObserver.add_entry(__file__, SetupA, SetupA.fixture_session, category="fixture", part="teardown",
                                  msg="begin execution TEARDOWN of fixture")

        self.SetupDevice1.s_i.do_something()
        self.SetupDevice1.s_ii.do_something()
        self.SetupDevice2.s_iii.do_something()
        self.SetupDevice2.s_iv.do_something()
        self.SetupUnmappedDevice1.s_v.do_something()
        self.SetupUnmappedDevice1.s_vi.do_something()
        self.SetupUnmappedDevice2.s_vii.do_something()
        self.SetupUnmappedDevice2.s_viii.do_something()

    @balder.fixture(level="setup")
    def fixture_setup(self):
        RuntimeObserver.add_entry(__file__, SetupA, SetupA.fixture_setup, category="fixture", part="construction",
                                  msg="begin execution CONSTRUCTION of fixture")

        self.SetupDevice1.s_i.do_something()
        self.SetupDevice1.s_ii.do_something()
        self.SetupDevice2.s_iii.do_something()
        self.SetupDevice2.s_iv.do_something()
        self.SetupUnmappedDevice1.s_v.do_something()
        self.SetupUnmappedDevice1.s_vi.do_something()
        self.SetupUnmappedDevice2.s_vii.do_something()
        self.SetupUnmappedDevice2.s_viii.do_something()

        yield

        RuntimeObserver.add_entry(__file__, SetupA, SetupA.fixture_setup, category="fixture", part="teardown",
                                  msg="begin execution TEARDOWN of fixture")

        self.SetupDevice1.s_i.do_something()
        self.SetupDevice1.s_ii.do_something()
        self.SetupDevice2.s_iii.do_something()
        self.SetupDevice2.s_iv.do_something()
        self.SetupUnmappedDevice1.s_v.do_something()
        self.SetupUnmappedDevice1.s_vi.do_something()
        self.SetupUnmappedDevice2.s_vii.do_something()
        self.SetupUnmappedDevice2.s_viii.do_something()

    @balder.fixture(level="scenario")
    def fixture_scenario(self):
        RuntimeObserver.add_entry(__file__, SetupA, SetupA.fixture_scenario, category="fixture", part="construction",
                                  msg="begin execution CONSTRUCTION of fixture")

        self.SetupDevice1.s_i.do_something()
        self.SetupDevice1.s_ii.do_something()
        self.SetupDevice2.s_iii.do_something()
        self.SetupDevice2.s_iv.do_something()
        self.SetupUnmappedDevice1.s_v.do_something()
        self.SetupUnmappedDevice1.s_vi.do_something()
        self.SetupUnmappedDevice2.s_vii.do_something()
        self.SetupUnmappedDevice2.s_viii.do_something()

        yield

        RuntimeObserver.add_entry(__file__, SetupA, SetupA.fixture_scenario, category="fixture", part="teardown",
                                  msg="begin execution TEARDOWN of fixture")

        self.SetupDevice1.s_i.do_something()
        self.SetupDevice1.s_ii.do_something()
        self.SetupDevice2.s_iii.do_something()
        self.SetupDevice2.s_iv.do_something()
        self.SetupUnmappedDevice1.s_v.do_something()
        self.SetupUnmappedDevice1.s_vi.do_something()
        self.SetupUnmappedDevice2.s_vii.do_something()
        self.SetupUnmappedDevice2.s_viii.do_something()

    @balder.fixture(level="variation")
    def fixture_variation(self):
        RuntimeObserver.add_entry(__file__, SetupA, SetupA.fixture_variation, category="fixture", part="construction",
                                  msg="begin execution CONSTRUCTION of fixture")

        self.SetupDevice1.s_i.do_something()
        self.SetupDevice1.s_ii.do_something()
        self.SetupDevice2.s_iii.do_something()
        self.SetupDevice2.s_iv.do_something()
        self.SetupUnmappedDevice1.s_v.do_something()
        self.SetupUnmappedDevice1.s_vi.do_something()
        self.SetupUnmappedDevice2.s_vii.do_something()
        self.SetupUnmappedDevice2.s_viii.do_something()

        yield

        RuntimeObserver.add_entry(__file__, SetupA, SetupA.fixture_variation, category="fixture", part="teardown",
                                  msg="begin execution TEARDOWN of fixture")

        self.SetupDevice1.s_i.do_something()
        self.SetupDevice1.s_ii.do_something()
        self.SetupDevice2.s_iii.do_something()
        self.SetupDevice2.s_iv.do_something()
        self.SetupUnmappedDevice1.s_v.do_something()
        self.SetupUnmappedDevice1.s_vi.do_something()
        self.SetupUnmappedDevice2.s_vii.do_something()
        self.SetupUnmappedDevice2.s_viii.do_something()

    @balder.fixture(level="testcase")
    def fixture_testcase(self):
        RuntimeObserver.add_entry(__file__, SetupA, SetupA.fixture_testcase, category="fixture", part="construction",
                                  msg="begin execution CONSTRUCTION of fixture")

        self.SetupDevice1.s_i.do_something()
        self.SetupDevice1.s_ii.do_something()
        self.SetupDevice2.s_iii.do_something()
        self.SetupDevice2.s_iv.do_something()
        self.SetupUnmappedDevice1.s_v.do_something()
        self.SetupUnmappedDevice1.s_vi.do_something()
        self.SetupUnmappedDevice2.s_vii.do_something()
        self.SetupUnmappedDevice2.s_viii.do_something()

        yield

        RuntimeObserver.add_entry(__file__, SetupA, SetupA.fixture_testcase, category="fixture", part="teardown",
                                  msg="begin execution TEARDOWN of fixture")

        self.SetupDevice1.s_i.do_something()
        self.SetupDevice1.s_ii.do_something()
        self.SetupDevice2.s_iii.do_something()
        self.SetupDevice2.s_iv.do_something()
        self.SetupUnmappedDevice1.s_v.do_something()
        self.SetupUnmappedDevice1.s_vi.do_something()
        self.SetupUnmappedDevice2.s_vii.do_something()
        self.SetupUnmappedDevice2.s_viii.do_something()



from __future__ import annotations
from typing import Union, List, Type, TYPE_CHECKING

from dataclasses import fields
from _balder.executor.setup_executor import SetupExecutor
from _balder.executor.basic_executable_executor import BasicExecutableExecutor
from _balder.fixture_execution_level import FixtureExecutionLevel
from _balder.testresult import ResultState, BranchBodyResult, ResultSummary
from _balder.previous_executor_mark import PreviousExecutorMark

if TYPE_CHECKING:
    from _balder.setup import Setup
    from _balder.fixture_manager import FixtureManager
    from _balder.executor.scenario_executor import ScenarioExecutor
    from _balder.executor.variation_executor import VariationExecutor
    from _balder.executor.testcase_executor import TestcaseExecutor


class ExecutorTree(BasicExecutableExecutor):
    """
    This class is the root object of the executor tree structure
    """
    fixture_execution_level = FixtureExecutionLevel.SESSION
    LINE_LENGTH = 120

    def __init__(self, fixture_manager: FixtureManager):
        super().__init__()
        self._setup_executors: List[SetupExecutor] = []
        self._fixture_manager = fixture_manager

        # contains the result object for the BODY part of this branch (will be overwritten in :class:`TestcaseExecutor`)
        self.body_result = BranchBodyResult(self)

    # ---------------------------------- STATIC METHODS ----------------------------------------------------------------

    # ---------------------------------- CLASS METHODS ----------------------------------------------------------------

    # ---------------------------------- PROPERTIES --------------------------------------------------------------------

    @property
    def all_child_executors(self) -> List[BasicExecutableExecutor] | None:
        return self._setup_executors

    @property
    def base_instance(self) -> object:
        """returns None because this element is a ExecutorTree"""
        return None

    @property
    def fixture_manager(self) -> FixtureManager:
        """returns the fixture manager of this tree"""
        return self._fixture_manager

    @property
    def parent_executor(self) -> None:
        return None

    # ---------------------------------- PROTECTED METHODS -------------------------------------------------------------

    def _prepare_execution(self, show_discarded):
        if not show_discarded:
            self.update_inner_feature_reference_in_all_setups()

    def _body_execution(self, show_discarded):
        for cur_setup_executor in self.get_setup_executors(return_discarded=show_discarded):
            prev_mark = cur_setup_executor.prev_mark
            if cur_setup_executor.has_runnable_tests(consider_discarded_too=show_discarded) \
                    or cur_setup_executor.has_skipped_tests():
                cur_setup_executor.execute(show_discarded=show_discarded)
            elif prev_mark == PreviousExecutorMark.SKIP:
                cur_setup_executor.set_result_for_whole_branch(ResultState.SKIP)
            elif prev_mark == PreviousExecutorMark.COVERED_BY:
                cur_setup_executor.set_result_for_whole_branch(ResultState.COVERED_BY)
            else:
                cur_setup_executor.set_result_for_whole_branch(ResultState.NOT_RUN)

    def _cleanup_execution(self, show_discarded):
        pass

    # ---------------------------------- METHODS -----------------------------------------------------------------------

    def get_setup_executors(self, return_discarded=False) -> List[SetupExecutor]:
        """
        returns all setup executors of this tree

        :param return_discarded: True if the method should return discarded variations too

        :return: a list of relevant :class:`SetupExecutor`
        """
        if return_discarded:
            return self._setup_executors

        return [
            cur_setup_executor
            for cur_setup_executor in self._setup_executors
            if len(cur_setup_executor.get_scenario_executors(return_discarded=False)) > 0
        ]

    def get_all_scenario_executors(self, return_discarded=False) -> List[ScenarioExecutor]:
        """
        returns a list with all scenario executors
        """
        all_scenario_executor = []
        for cur_setup_executor in self.get_setup_executors(return_discarded=return_discarded):
            all_scenario_executor += cur_setup_executor.get_scenario_executors(return_discarded=return_discarded)
        return all_scenario_executor

    def get_all_variation_executors(self, return_discarded=False) -> List[VariationExecutor]:
        """
        returns a list with all variation executors
        """
        all_variation_executor = []
        for cur_scenario_executor in self.get_all_scenario_executors(return_discarded=return_discarded):
            all_variation_executor += cur_scenario_executor.get_variation_executors(return_discarded=return_discarded)
        return all_variation_executor

    def get_all_testcase_executors(self) -> List[TestcaseExecutor]:
        """
        returns a list with all testcase executors
        """
        all_testcase_executor = []
        for cur_scenario_executor in self.get_all_scenario_executors():
            for cur_variation_executor in cur_scenario_executor.get_variation_executors():
                all_testcase_executor += cur_variation_executor.get_testcase_executors()
        return all_testcase_executor

    def add_setup_executor(self, setup_executor: SetupExecutor):
        """
        This method adds a new :class:`SetupExecutor` to the child element's list of this tree object
        """
        if not isinstance(setup_executor, SetupExecutor):
            raise TypeError("the given object `setup_executor` must be of type `SetupExecutor`")
        if setup_executor in self._setup_executors:
            raise ValueError("the given object `setup_executor` already exists in child list")
        self._setup_executors.append(setup_executor)

    def get_executor_for_setup(self, setup: Type[Setup]) -> Union[SetupExecutor, None]:
        """
        This method searches for a SetupExecutor in the internal list for the given :class:`Setup` type

        :param setup: the setup class for which the executor is being searched for

        :return: returns the associated SetupExecutor or None if there are no matches for the given type
        """
        for cur_setup_executor in self._setup_executors:
            if cur_setup_executor.base_setup_class.__class__ == setup:
                return cur_setup_executor
        # can not find some
        return None

    def cleanup_empty_executor_branches(self, consider_discarded=False):
        to_remove_executor = []
        for cur_setup_executor in self.get_setup_executors(return_discarded=consider_discarded):
            cur_setup_executor.cleanup_empty_executor_branches(consider_discarded=consider_discarded)
            if len(cur_setup_executor.get_scenario_executors(return_discarded=consider_discarded)) == 0:
                # remove this whole executor because it has no children anymore
                to_remove_executor.append(cur_setup_executor)
        for cur_setup_executor in to_remove_executor:
            self._setup_executors.remove(cur_setup_executor)

    def update_inner_feature_reference_in_all_setups(self):
        """
        This method iterates over all setups and triggers the method that updates the inner feature references.

        This needs to be done on ExecutorTree level, because this reference should also be accessible within session
        fixtures (in case they have setup scope).
        """
        for cur_setup_executor in self.get_setup_executors():
            cur_setup_executor.update_inner_referenced_feature_instances()

    def execute(self, show_discarded=False) -> None:
        """
        This method executes this branch of the tree
        """
        start_text = "START TESTSESSION"
        end_text = "FINISH TESTSESSION"

        def print_line(text):
            full_text = int((self.LINE_LENGTH - (len(start_text) + 2)) / 2) * "=" + " " + text + " "
            full_text += "=" * (self.LINE_LENGTH - len(full_text))
            print(full_text)

        print_line(start_text)
        # check if there exists runnable elements
        runnables = [cur_exec.has_runnable_tests(consider_discarded_too=show_discarded)
                     for cur_exec in self.get_setup_executors(return_discarded=show_discarded)]
        one_or_more_runnable_setups = None if len(runnables) == 0 else max(runnables)
        if one_or_more_runnable_setups:
            super().execute(show_discarded=show_discarded)
        else:
            print("NO EXECUTABLE SETUPS/SCENARIOS FOUND")
        print_line(end_text)
        summary = self.testsummary()
        is_first = True
        for cur_field in fields(ResultSummary):
            if is_first:
                is_first = False
            else:
                print(" | ", end="")
            print(f"TOTAL {cur_field.name.upper()}: {getattr(summary, cur_field.name)}", end="")
        print("")

    def print_tree(self, show_discarded=False) -> None:
        """this method is an auxiliary method which outputs the entire tree"""
        print("RESOLVING OVERVIEW", end="\n\n")
        for cur_setup_executor in self.get_setup_executors(return_discarded=show_discarded):
            for cur_scenario_executor in cur_setup_executor.get_scenario_executors(return_discarded=show_discarded):
                for cur_variation_executor in cur_scenario_executor.get_variation_executors(
                        return_discarded=show_discarded):
                    applicable = cur_variation_executor.prev_mark != PreviousExecutorMark.DISCARDED
                    start_char = '+' if applicable else 'X'
                    print(start_char * self.LINE_LENGTH)
                    applicability_str = "[APPLICABLE]" if applicable else "[DISCARDED] "
                    print(f"{start_char} {applicability_str} Scenario "
                          f"`{cur_scenario_executor.base_scenario_class.__class__.__qualname__}` <-> "
                          f"Setup `{cur_setup_executor.base_setup_class.__class__.__qualname__}`")
                    mapping_printings = {}
                    for cur_scenario_device, cur_setup_device in cur_variation_executor.base_device_mapping.items():
                        mapping_printings[f"   {cur_scenario_device.__qualname__}"] = str(cur_setup_device.__qualname__)
                    max_len = max(len(cur_elem) for cur_elem in mapping_printings.keys())
                    for cur_key, cur_val in mapping_printings.items():
                        print(("{} {:<" + str(max_len) + "} = {}").format(start_char, cur_key, cur_val))
                    for cur_testcase_excutor in cur_variation_executor.get_testcase_executors():
                        print(f"{start_char}    -> Testcase<"
                              f"{cur_testcase_excutor.base_testcase_callable.__qualname__}>")
                    if cur_variation_executor.prev_mark == PreviousExecutorMark.DISCARDED:
                        print(f"{start_char}")
                        print(f"{start_char}    DISCARDED BECAUSE "
                              f"`{cur_variation_executor.not_applicable_variation_exc.args[0]}`")
                    print(('+' if applicable else 'X') * self.LINE_LENGTH)
                    print('')



from __future__ import annotations

import inspect
from typing import List, Type, Any, Dict, Iterable, TYPE_CHECKING
import types
from graphlib import TopologicalSorter
from collections import OrderedDict

from _balder.executor.basic_executor import BasicExecutor
from _balder.executor.parametrized_testcase_executor import ParametrizedTestcaseExecutor
from _balder.parametrization import Parameter
from _balder.previous_executor_mark import PreviousExecutorMark
from _balder.testresult import BranchBodyResult
from _balder.utils.mixin_can_be_covered_by_executor import MixinCanBeCoveredByExecutor

if TYPE_CHECKING:
    from _balder.executor.scenario_executor import ScenarioExecutor
    from _balder.executor.variation_executor import VariationExecutor
    from _balder.scenario import Scenario
    from _balder.setup import Setup


class UnresolvedParametrizedTestcaseExecutor(BasicExecutor, MixinCanBeCoveredByExecutor):
    """
    This executor class represents a group of dynamically parametrized tests.
    """

    def __init__(
            self,
            testcase: callable,
            parent: VariationExecutor,
            static_parameters: Dict[str, Any] = None,
    ) -> None:
        super().__init__()

        self._base_testcase_callable = testcase
        self._parent_executor = parent

        # holds the specific static parameters for this unresolved group
        self._static_parameters = static_parameters if static_parameters is not None else {}

        # contains the result object for the BODY part of this branch
        self.body_result = BranchBodyResult(self)

    @property
    def parent_executor(self) -> VariationExecutor:
        return self._parent_executor

    @property
    def scenario_executor(self) -> ScenarioExecutor:
        return self.parent_executor.parent_executor

    @property
    def base_testcase_callable(self) -> callable:
        """returns the testcase function"""
        return self._base_testcase_callable

    @property
    def all_child_executors(self) -> None:
        return None

    @property
    def base_testcase_obj(self) -> Scenario:
        """
        return the testcase Scenario this testcase is defined in
        """
        return self.parent_executor.cur_scenario_class

    @property
    def base_instance(self) -> object:
        """
        returns the base class instance to which this executor instance belongs"""
        return self._base_testcase_callable

    def has_skipped_tests(self) -> bool:
        return self.prev_mark == PreviousExecutorMark.SKIP

    def has_covered_by_tests(self) -> bool:
        return self.prev_mark == PreviousExecutorMark.COVERED_BY

    def get_all_base_instances_of_this_branch(
            self, with_type: Type[Setup] | Type[Scenario] | Type[types.FunctionType],
            only_runnable_elements: bool = True) -> List[Setup | Scenario | object]:
        """
        returns itself if the type matches, otherwise an empty list
        """
        # todo
        if isinstance(self.base_instance, with_type):
            return [self.base_instance]
        return []

    def get_covered_by_element(self) -> List[Scenario | callable]:
        covered_by_dict = self.scenario_executor.base_scenario_controller.get_abs_covered_by_dict()
        all_covered_by_data = covered_by_dict.get(self.base_testcase_callable.__name__, [])
        # also add all scenario specified covered-by elements
        all_covered_by_data.extend(covered_by_dict.get(None, []))
        return all_covered_by_data

    def get_resolved_parametrized_testcase_executors(self):
        """
        resolves the dynamic parametrization - should be called when setup features are active in the scenario
        """
        executors = []

        parametrization = self.get_parametrization()

        if not parametrization:
            return []

        for cur_parametrization in parametrization:
            executors.append(
                ParametrizedTestcaseExecutor(
                    self._base_testcase_callable,
                    parent=self.parent_executor,
                    parametrization=cur_parametrization,
                    unresolved_group_obj=self
                )
            )
        return executors

    def get_parametrization(self) -> List[OrderedDict[str, Any]] | None:
        """
        returns all parametrization elements that belongs to this group executor
        """
        scenario_controller = self.parent_executor.parent_executor.base_scenario_controller
        dynamic_parametrization = scenario_controller.get_parametrization_for(self._base_testcase_callable,
                                                                              static=False)
        if not dynamic_parametrization:
            raise ValueError('can not determine dynamic parametrization, because there are no dynamic parameters')

        # sort attributes according their Parameter - using TopologicalSorter
        graph = {attr: [param.name for param in config.get_parameters(of_type=Parameter).values()]
                 for attr, config in dynamic_parametrization.items()}
        # also add all elements from static parameters
        graph.update({param: [] for param in self._static_parameters})

        ts = TopologicalSorter(graph)
        resolvable_order_of_attribues = ts.static_order()
        resolvable_dynamic_attribues = [attr for attr in resolvable_order_of_attribues
                                        if attr in dynamic_parametrization.keys()]

        def get_variations_for(
                resolved_parameters: Dict[str, Any],
                remaining_attributes: Iterable[str]
        ) -> List[Dict[str, Any]]:
            result = []
            remaining_attributes = list(remaining_attributes).copy()
            attr = remaining_attributes.pop(0)
            feature_access_selector = dynamic_parametrization[attr]
            # get value for this attribute
            attr_value_list = feature_access_selector.get_value(resolved_parameters)
            if not isinstance(attr_value_list, Iterable):
                raise TypeError(
                    f'feature parametrizing not possible, because `{feature_access_selector.device.__qualname__}'
                    f'.{feature_access_selector.device_property_name}.{feature_access_selector.feature_property_name}` '
                    f'does not return Iterable')

            for cur_value in attr_value_list:
                parameters_with_cur_attr_value = resolved_parameters.copy()
                parameters_with_cur_attr_value[attr] = cur_value
                if len(remaining_attributes) > 0:
                    result.extend(get_variations_for(parameters_with_cur_attr_value, remaining_attributes))
                else:
                    result.append(parameters_with_cur_attr_value)

            return result

        resolved_parameters = self._static_parameters.copy()

        all_full_parametrization = get_variations_for(resolved_parameters, resolvable_dynamic_attribues)

        # get combined parametrization
        result = []
        for cur_full_parametrization in all_full_parametrization:
            cur_parameter_set = OrderedDict()
            for cur_arg in inspect.getfullargspec(self._base_testcase_callable).args:
                # only if this is a parametrization value
                if cur_arg in cur_full_parametrization:
                    cur_parameter_set[cur_arg] = cur_full_parametrization[cur_arg]
            result.append(cur_parameter_set)
        return result

    def cleanup_empty_executor_branches(self, consider_discarded=False):
        pass



import balder


class ProvidesANumberFeature(balder.Feature):

    the_number = 0

    def set_number(self, number):
        """user method that allows to set the number"""
        self.the_number = number

    def get_number(self):
        """returns the set number"""
        raise NotImplementedError("has to be implemented in subclass")

    def sends_the_number(self):
        """sends the set number"""
        raise NotImplementedError("has to be implemented in subclass")


class AddCalculateFeature(balder.Feature):
    """
    note: Although this feature has two vdevices that define a number provider, please note that multiple vdevice
    mappings are not yet supported.
    """

    all_numbers = []

    class ANumberProviderDevice1(balder.VDevice):
        provider = ProvidesANumberFeature()

    class ANumberProviderDevice2(balder.VDevice):
        provider = ProvidesANumberFeature()

    def get_numbers(self):
        """this method get all the single numbers that should be calculated"""
        raise NotImplementedError("has to be implemented in subclass")

    def add_numbers(self):
        """this method adds all the numbers"""
        raise NotImplementedError("has to be implemented in subclass")

    def get_number_of_provider_1(self):
        return self.ANumberProviderDevice1.provider.get_number()

    def get_number_of_provider_2(self):
        return self.ANumberProviderDevice2.provider.get_number()



import balder
import logging
from .setup_features import SetupFeatureI, SetupFeatureII
from ..lib.connections import AConnection
from ..lib.utils import FixtureReturn
from ..balderglob import RuntimeObserver

logger = logging.getLogger(__name__)


class SetupA(balder.Setup):
    """This is a setup of category A (exactly the same as scenario A)"""

    class SetupDevice1(balder.Device):
        s_i = SetupFeatureI()

    @balder.connect(SetupDevice1, over_connection=AConnection)
    class SetupDevice2(balder.Device):
        s_ii = SetupFeatureII()

    @balder.fixture(level="session")
    def fixture_session(self, fixture_session):
        RuntimeObserver.add_entry(__file__, SetupA, SetupA.fixture_session, category="fixture", part="construction",
                                  msg="begin execution CONSTRUCTION of fixture")

        self.SetupDevice1.s_i.do_something()
        self.SetupDevice2.s_ii.do_something()

        assert isinstance(self, SetupA)
        assert fixture_session == FixtureReturn.BALDERGLOB_SESSION

        yield FixtureReturn.SETUP_A_SESSION

        RuntimeObserver.add_entry(__file__, SetupA, SetupA.fixture_session, category="fixture", part="teardown",
                                  msg="begin execution TEARDOWN of fixture")

        self.SetupDevice1.s_i.do_something()
        self.SetupDevice2.s_ii.do_something()

    @balder.fixture(level="setup")
    def fixture_setup(self, fixture_session, fixture_setup):
        RuntimeObserver.add_entry(__file__, SetupA, SetupA.fixture_setup, category="fixture", part="construction",
                                  msg="begin execution CONSTRUCTION of fixture")

        self.SetupDevice1.s_i.do_something()
        self.SetupDevice2.s_ii.do_something()

        assert isinstance(self, SetupA)
        assert fixture_session == FixtureReturn.SETUP_A_SESSION
        assert fixture_setup == FixtureReturn.BALDERGLOB_SETUP

        yield FixtureReturn.SETUP_A_SETUP

        RuntimeObserver.add_entry(__file__, SetupA, SetupA.fixture_setup, category="fixture", part="teardown",
                                  msg="begin execution TEARDOWN of fixture")

        self.SetupDevice1.s_i.do_something()
        self.SetupDevice2.s_ii.do_something()

    @balder.fixture(level="scenario")
    def fixture_scenario(self, fixture_session, fixture_setup, fixture_scenario):
        RuntimeObserver.add_entry(__file__, SetupA, SetupA.fixture_scenario, category="fixture", part="construction",
                                  msg="begin execution CONSTRUCTION of fixture")

        self.SetupDevice1.s_i.do_something()
        self.SetupDevice2.s_ii.do_something()

        assert isinstance(self, SetupA)
        assert fixture_session == FixtureReturn.SETUP_A_SESSION
        assert fixture_setup == FixtureReturn.SETUP_A_SETUP
        assert fixture_scenario == FixtureReturn.BALDERGLOB_SCENARIO

        yield FixtureReturn.SETUP_A_SCENARIO

        RuntimeObserver.add_entry(__file__, SetupA, SetupA.fixture_scenario, category="fixture", part="teardown",
                                  msg="begin execution TEARDOWN of fixture")

        self.SetupDevice1.s_i.do_something()
        self.SetupDevice2.s_ii.do_something()

    @balder.fixture(level="variation")
    def fixture_variation(self, fixture_session, fixture_setup, fixture_scenario, fixture_variation):
        RuntimeObserver.add_entry(__file__, SetupA, SetupA.fixture_variation, category="fixture", part="construction",
                                  msg="begin execution CONSTRUCTION of fixture")

        self.SetupDevice1.s_i.do_something()
        self.SetupDevice2.s_ii.do_something()

        assert isinstance(self, SetupA)
        assert fixture_session == FixtureReturn.SETUP_A_SESSION
        assert fixture_setup == FixtureReturn.SETUP_A_SETUP
        assert fixture_scenario == FixtureReturn.SETUP_A_SCENARIO
        assert fixture_variation == FixtureReturn.BALDERGLOB_VARIATION

        yield FixtureReturn.SETUP_A_VARIATION

        RuntimeObserver.add_entry(__file__, SetupA, SetupA.fixture_variation, category="fixture", part="teardown",
                                  msg="begin execution TEARDOWN of fixture")

        self.SetupDevice1.s_i.do_something()
        self.SetupDevice2.s_ii.do_something()

    @balder.fixture(level="testcase")
    def fixture_testcase(self, fixture_session, fixture_setup, fixture_scenario, fixture_variation, fixture_testcase):
        RuntimeObserver.add_entry(__file__, SetupA, SetupA.fixture_testcase, category="fixture", part="construction",
                                  msg="begin execution CONSTRUCTION of fixture")

        self.SetupDevice1.s_i.do_something()
        self.SetupDevice2.s_ii.do_something()

        assert isinstance(self, SetupA)
        assert fixture_session == FixtureReturn.SETUP_A_SESSION
        assert fixture_setup == FixtureReturn.SETUP_A_SETUP
        assert fixture_scenario == FixtureReturn.SETUP_A_SCENARIO
        assert fixture_variation == FixtureReturn.SETUP_A_VARIATION
        assert fixture_testcase == FixtureReturn.BALDERGLOB_TESTCASE

        yield FixtureReturn.SETUP_A_TESTCASE

        RuntimeObserver.add_entry(__file__, SetupA, SetupA.fixture_testcase, category="fixture", part="teardown",
                                  msg="begin execution TEARDOWN of fixture")

        self.SetupDevice1.s_i.do_something()
        self.SetupDevice2.s_ii.do_something()



import balder


@balder.insert_into_tree()
class ParentConnection(balder.Connection):
    pass


@balder.insert_into_tree(parents=[ParentConnection])
class ChildAConnection(balder.Connection):
    pass


@balder.insert_into_tree(parents=[ParentConnection])
class ChildBConnection(balder.Connection):
    pass



import balder
import logging
from .setup_features import SetupFeatureIII, SetupFeatureIV
from ..lib.connections import ChildConnection, ParentConnection

logger = logging.getLogger(__name__)


class SetupB(balder.Setup):
    """This is a setup of category B (exactly the same as scenario B)"""

    class SetupDevice3(balder.Device):
        iii = SetupFeatureIII()

    @balder.connect(SetupDevice3, over_connection=ChildConnection.based_on(ParentConnection))
    class SetupDevice4(balder.Device):
        iv = SetupFeatureIV()



from __future__ import annotations
from typing import Type, Dict, Any, TypeVar, List
import dataclasses

from _balder.device import Device

ValueTypeT = TypeVar("ValueTypeT")


@dataclasses.dataclass
class FeatureAccessSelector:
    """helper object for dynamic parametrizing by feature method/property"""
    device: Type[Device]
    device_property_name: str
    feature_property_name: str
    parameters: Dict[str, FeatureAccessSelector | Value] = dataclasses.field(default_factory=dict)

    def get_value(self, available_parameters: Dict[str, Any]) -> List[Any]:
        """accesses the configured method/property"""
        resolved_parameters = {}
        for cur_key, cur_value in self.parameters.items():
            if isinstance(cur_value, FeatureAccessSelector):
                resolved_parameters[cur_key] = cur_value.get_value(available_parameters)
            elif isinstance(cur_value, Parameter):
                resolved_parameters[cur_key] = available_parameters[cur_value.name]
            elif isinstance(cur_value, Value):
                resolved_parameters[cur_key] = cur_value.value
            else:
                raise TypeError(f'go unexpected parameter type {type(cur_value)} in FeatureAccessSelector')

        feature = getattr(self.device, self.device_property_name)

        if isinstance(getattr(feature.__class__, self.feature_property_name), property):
            return getattr(feature, self.feature_property_name)

        return getattr(feature, self.feature_property_name)(**resolved_parameters)

    def get_parameters(
            self,
            of_type: Type[FeatureAccessSelector | Parameter | Value] | None = None
    ) -> Dict[str, FeatureAccessSelector | Parameter | Value]:
        """
        Returns the parameters of this access selector

        :param of_type: allows to filter the parameters by the value's type
        """
        result = {}
        for cur_attr, cur_value in self.parameters.items():
            if of_type is None or isinstance(cur_value, of_type):
                result[cur_attr] = cur_value
        return result


class Parameter:
    """allows to parametrize a parametrization by another parametrization value"""

    def __init__(self, name: str) -> None:
        self._name = name

    @property
    def name(self) -> str:
        """returns the name of the parameter"""
        return self._name


class Value:
    """allows to parametrize a parametrization by a fix value"""

    def __init__(self, value: ValueTypeT) -> None:
        self._value = value

    @property
    def value(self) -> ValueTypeT:
        """returns the value of the parametrization"""
        return self._value



from __future__ import annotations
from enum import Enum


class PreviousExecutorMark(Enum):
    """
    This enum defines the previous state of an executor object. It allows to set a pre-state to an executor, that
    describes if it is executable before the executor-tree will be called
    """
    #: this is the normal state, where all sub elements are allowed to be executed
    RUNNABLE = 1

    #: this marks the element and all of its sub elements as skip-able
    SKIP = 2

    #: this marks the element and all of its sub elements as NOT RUNNABLE, because this element is covered_by another
    COVERED_BY = 3

    #: this marks that the element should be ignored from the executor mechanism
    IGNORE = -1

    #: this marks that the element was already discarded because its variation is not applicable
    DISCARDED = -2



from __future__ import annotations
from typing import List, Union, TYPE_CHECKING

import sys
import time
import traceback

from _balder.executor.basic_executable_executor import BasicExecutableExecutor
from _balder.fixture_execution_level import FixtureExecutionLevel
from _balder.previous_executor_mark import PreviousExecutorMark
from _balder.testresult import ResultState, TestcaseResult
from _balder.utils.functions import get_method_type
from _balder.utils.mixin_can_be_covered_by_executor import MixinCanBeCoveredByExecutor

if TYPE_CHECKING:
    from _balder.executor.scenario_executor import ScenarioExecutor
    from _balder.executor.variation_executor import VariationExecutor
    from _balder.fixture_manager import FixtureManager
    from _balder.scenario import Scenario


class TestcaseExecutor(BasicExecutableExecutor, MixinCanBeCoveredByExecutor):
    """
    A TestcaseExecutor class represents an actual single test that can be executed. It therefore references exactly to a
    test method of a scenario that can be executed on the specific setup this executor belongs to.
    """
    fixture_execution_level = FixtureExecutionLevel.TESTCASE

    def __init__(
            self,
            testcase: callable,
            parent: VariationExecutor,
    ) -> None:
        super().__init__()

        self._base_testcase_callable = testcase
        self._parent_executor = parent
        self._fixture_manager: FixtureManager = parent.fixture_manager

        # if there exist executors that cover this item, they are contained as list in this property
        self.covered_by_executors = None

        # contains the result object for the BODY part of this branch
        self.body_result = TestcaseResult(self)

        # holds the raw test execution time in seconds
        self.test_execution_time_sec = 0

        # determine prev_mark IGNORE/SKIP for the testcase
        if self.should_be_skipped():
            self.prev_mark = PreviousExecutorMark.SKIP
        # always overwrite if it should be ignored
        if self.should_be_ignored():
            self.prev_mark = PreviousExecutorMark.IGNORE

    # ---------------------------------- STATIC METHODS ----------------------------------------------------------------

    # ---------------------------------- CLASS METHODS ----------------------------------------------------------------

    # ---------------------------------- PROPERTIES --------------------------------------------------------------------

    @property
    def all_child_executors(self) -> None:
        return None

    @property
    def parent_executor(self) -> VariationExecutor:
        return self._parent_executor

    @property
    def scenario_executor(self) -> ScenarioExecutor:
        return self.parent_executor.parent_executor

    @property
    def base_instance(self) -> object:
        """
        returns the base class instance to which this executor instance belongs"""
        return self._base_testcase_callable

    @property
    def base_testcase_callable(self) -> callable:
        """returns the testcase function"""
        return self._base_testcase_callable

    @property
    def base_testcase_obj(self) -> Scenario:
        """
        return the testcase Scenario this testcase is defined in
        """
        return self._parent_executor.cur_scenario_class

    @property
    def fixture_manager(self):
        """returns the current active fixture manager"""
        return self._fixture_manager

    @property
    def full_test_name_str(self) -> str:
        """
        returns the name of the test method, that should be used in output
        """
        return self._base_testcase_callable.__qualname__

    # ---------------------------------- PROTECTED METHODS -------------------------------------------------------------

    def _prepare_execution(self, show_discarded):
        print(f"      TEST {self.full_test_name_str} ", end='')

    def _body_execution(self, show_discarded):
        self.test_execution_time_sec = 0

        if self.should_be_skipped():
            self.body_result.set_result(ResultState.SKIP)
            return
        if self.should_be_ignored():
            self.body_result.set_result(ResultState.NOT_RUN)
            return
        if self.is_covered_by():
            self.body_result.set_result(ResultState.COVERED_BY)
            return

        start_time = time.perf_counter()
        try:
            func_type = get_method_type(self.base_testcase_obj.__class__, self.base_testcase_callable)
            all_args = self.get_all_test_method_args()
            if func_type == "staticmethod":
                # testcase is a staticmethod - no special first attribute
                self.base_testcase_callable(**all_args)
            elif func_type == "classmethod":
                self.base_testcase_callable(self=self.base_testcase_obj.__class__, **all_args)
            elif func_type == "instancemethod":
                self.base_testcase_callable(self=self.base_testcase_obj, **all_args)
            else:
                # `function` is not allowed here!
                raise ValueError(f"found illegal value for func_type `{func_type}` for test "
                                 f"`{self.base_testcase_callable.__name__}`")

            self.body_result.set_result(ResultState.SUCCESS)
        except Exception as exc:  # pylint: disable=broad-exception-caught
            # this has to be a test error
            traceback.print_exception(*sys.exc_info())
            self.body_result.set_result(ResultState.FAILURE, exc)
        self.test_execution_time_sec = time.perf_counter() - start_time

    def _cleanup_execution(self, show_discarded):
        print(f"[{self.body_result.get_result_as_char()}]")

    # ---------------------------------- METHODS -----------------------------------------------------------------------

    def should_run(self):
        """returns true if the testcase should be executed (defined in scenario)"""
        if self.base_testcase_callable in self.parent_executor.parent_executor.all_run_tests:
            return True
        return False

    def should_be_skipped(self):
        """returns true if the testcase should be skipped (defined in scenario)"""
        if self.base_testcase_callable in self.parent_executor.parent_executor.all_skip_tests:
            return True
        return False

    def should_be_ignored(self):
        """returns true if the testcase should be ignored (defined in scenario)"""
        if self.base_testcase_callable in self.parent_executor.parent_executor.all_ignore_tests:
            return True
        return False

    def is_covered_by(self):
        """returns true if the testcase is covered-by"""
        return self.prev_mark == PreviousExecutorMark.COVERED_BY

    def has_skipped_tests(self) -> bool:
        return self.prev_mark == PreviousExecutorMark.SKIP

    def has_covered_by_tests(self) -> bool:
        return self.prev_mark == PreviousExecutorMark.COVERED_BY

    def cleanup_empty_executor_branches(self, consider_discarded=False):
        """
        This method searches the whole tree and removes branches where an executor item has no own children. It can
        remove these branches, because they have no valid matchings.

        This method implementation of the :class:`TestcaseExecutor` does nothing.
        """

    def get_covered_by_element(self) -> List[Union[Scenario, callable]]:
        covered_by_dict = self.scenario_executor.base_scenario_controller.get_abs_covered_by_dict()
        all_covered_by_data = covered_by_dict.get(self.base_testcase_callable.__name__, [])
        # also add all scenario specified covered-by elements
        all_covered_by_data.extend(covered_by_dict.get(None, []))
        return all_covered_by_data

    def get_all_test_method_args(self):
        """
        returns all kwargs values for the test method
        """
        func_type = get_method_type(self.base_testcase_obj.__class__, self.base_testcase_callable)
        return self.fixture_manager.get_all_attribute_values(
            self,
            self.base_testcase_obj.__class__,
            self.base_testcase_callable,
            func_type
        )



import balder
from .features_setup import SendOverLocalVarFeature, ReceiveOverLocalVarFeature


class SetupVarAliceToBob(balder.Setup):
    class Alice(balder.Device):
        send = SendOverLocalVarFeature()

    @balder.connect(with_device=Alice, over_connection=balder.Connection())
    class Bob(balder.Device):
        recv = ReceiveOverLocalVarFeature()



from __future__ import annotations
from typing import TYPE_CHECKING, List, Type, Tuple, Union

import argparse
import pathlib
from _balder.balder_plugin import BalderPlugin
from _balder.exceptions import UnexpectedPluginMethodReturnValue

if TYPE_CHECKING:
    from _balder.balder_session import BalderSession
    from _balder.executor.executor_tree import ExecutorTree
    from _balder.scenario import Scenario
    from _balder.setup import Setup


class PluginManager:
    """
    This class is a helper class that manages all existing plugins.
    """

    def __init__(self):
        self.all_plugins: List[BalderPlugin] = []

    def register(self, plugin_class: Type[BalderPlugin], session: BalderSession):
        """
        This method registers a new plugin class.

        :param plugin_class: the new class that should be registered

        :param session: the balder session, the new plugin belongs to
        """
        if plugin_class in [cur_plugin_instance.__class__ for cur_plugin_instance in self.all_plugins]:
            raise ValueError(f"the given plugin class `{plugin_class.__name__}` can not be added twice")
        plugin_instance = plugin_class(session)
        self.all_plugins.append(plugin_instance)

    def execute_addoption(self, argument_parser: argparse.ArgumentParser) -> None:
        """
        This method executes all plugin callbacks :ref:`BalderPlugin.addoption`.

        The callback will be executed while the `ArgumentParser` is being created.

        :param argument_parser: the argument parser object
        """
        for cur_plugin in self.all_plugins:
            cur_plugin.addoption(argument_parser)

    def execute_modify_collected_pyfiles(self, pyfiles: List[pathlib.Path]) -> List[pathlib.Path]:
        """
        This method executes all plugin callbacks :ref:`BalderPlugin.cb_filter_executor_tree`.

        The callback will be executed to filter the collected pyfiles, that will be analysed later. From these files,
        balder tries to determine and then import relevant :class:`Scenario` and :class:`Setup` classes later.

        :param pyfiles: a list with all collected pyfiles

        :return: the new list that should be used for the pyfiles
        """
        cur_pyfiles = pyfiles
        for cur_plugin in self.all_plugins:
            if cur_plugin.__class__.modify_collected_pyfiles == BalderPlugin.modify_collected_pyfiles:
                # there is no specific implementation
                break
            ret = cur_plugin.modify_collected_pyfiles(pyfiles=cur_pyfiles)
            if ret is None:
                raise UnexpectedPluginMethodReturnValue(
                    f"the method `modify_collected_pyfiles` of balder plugin `{cur_plugin.__class__.__name__}` has not"
                    f"returned something - a list of filepaths was expected")
            if not isinstance(ret, list):
                raise UnexpectedPluginMethodReturnValue(
                    f"the method `modify_collected_pyfiles` of balder plugin `{cur_plugin.__class__.__name__}` has not"
                    f"returned a list (return type is `{type(ret)}`)")
            cur_pyfiles = ret
        return cur_pyfiles

    def execute_collected_classes(self, scenarios: List[Type[Scenario]], setups: List[Type[Setup]]) \
            -> Tuple[List[Type[Scenario]], List[Type[Setup]]]:
        """
        This method executes all plugin callbacks :ref:`BalderPlugin.cb_filter_executor_tree`.

        The callback will be executed to filter the collected :class:`Scenario` and :class:`Setup` classes.

        :param scenarios: a list with all collected :class:`Scenario` classes before the plugin can filter it

        :param setups: a list with all collected :class:`Setup` classes before the plugin can filter it

        :return: a tuple where the first element is the new :class:`Scenario` list and the second is the new
                 :class:`Setup` class list
        """
        cur_scenarios = scenarios
        cur_setups = setups
        for cur_plugin in self.all_plugins:
            if cur_plugin.__class__.modify_collected_classes == BalderPlugin.modify_collected_classes:
                # there is no specific implementation
                break
            ret = cur_plugin.modify_collected_classes(scenarios=cur_scenarios, setups=cur_setups)
            if ret is None:
                raise UnexpectedPluginMethodReturnValue(
                    f"the method `execute_collected_classes` of balder plugin `{cur_plugin.__class__.__name__}` has not"
                    f"returned something - a tuple with two lists for Scenarios and Setups was expected")
            if not isinstance(ret, tuple):
                raise UnexpectedPluginMethodReturnValue(
                    f"the method `execute_collected_classes` of balder plugin `{cur_plugin.__class__.__name__}` has not"
                    f"returned a tuple (return type is `{type(ret)}`)")
            if not len(ret) == 2:
                raise UnexpectedPluginMethodReturnValue(
                    f"the method `execute_collected_classes` of balder plugin `{cur_plugin.__class__.__name__}` has not"
                    f"returned a tuple of length 2")
            if not isinstance(ret[0], list) or not isinstance(ret[1], list):
                raise UnexpectedPluginMethodReturnValue(
                    f"the method `execute_collected_classes` of balder plugin `{cur_plugin.__class__.__name__}` has not"
                    f"returned a tuple with two lists as values")
            cur_scenarios, cur_setups = ret
        return cur_scenarios, cur_setups

    def execute_filter_executor_tree(self, executor_tree: ExecutorTree) -> None:
        """
        This method executes all plugin methods :ref:`BalderPlugin.filter_executor_tree`.

        The callback will be executed to filter the created :class:`ExecutorTree` for this balder session. For this the
        :class:`ExecutorTree` reference (given by ``executor_tree``) can be manipulated.

        :param executor_tree: the reference to the main :class:`ExecutorTree` object that balder uses for this session
        """
        for cur_plugin in self.all_plugins:
            cur_plugin.filter_executor_tree(executor_tree=executor_tree)

    def execute_session_finished(self, executor_tree: Union[ExecutorTree, None]) -> None:
        """
        This method executes all plugin methods :ref:`BalderPlugin.session_finished`.

        This callback will be executed at the end of every session. The callback will run in a `collect-only` and
        `resolve-only` session too. Note, that the `executor_tree` argument is None in a `collect-only` session.

        :param executor_tree: the reference to the main :class:`ExecutorTree` object that balder uses for this session
        """
        for cur_plugin in self.all_plugins:
            cur_plugin.session_finished(executor_tree=executor_tree)



import balder


class ScenarioNotMatching(balder.Scenario):
    """This scenario should not be collected"""

    class MyDevice(balder.Device):
        pass
    
    def test_do_something(self):
        assert True



from __future__ import annotations


class BalderException(Exception):
    """basic balder exception"""


class BalderWarning(Warning):
    """basic balder warning"""


class FixtureScopeError(BalderException):
    """is thrown for fixtures that were defined in an illegal scope"""


class FixtureReferenceError(BalderException):
    """
    is thrown when an error is detected in the reference between fixtures
    """


class UnclearSetupScopedFixtureReference(BalderException):
    """
    is thrown for the special UNCLEAR SETUP SCOPED FIXTURE REFERENCE error
    """


class UnclearUniqueClassReference(BalderException):
    """
    is thrown after balder can not clearly determine the related instance reference for a :class:`Scenario` or
    :class:`Setup` class
    """


class LostInExecutorTreeException(BalderException):
    """
    is thrown if you make a wrong call when entering and leaving the fixture manager, and you got lost
    """


class DeviceResolvingException(BalderException):
    """
    is thrown if an error occurs while resolving a device string
    """


class NodeNotExistsError(BalderException):
    """
    is thrown if no connection node exists for a device
    """


class DuplicateForVDeviceError(BalderException):
    """
    is thrown if some illegal `@for_vdevice` decorator exists for a feature
    """


class DuplicateBalderSettingError(BalderException):
    """
    is thrown if balder can not determine the BalderSettings object clearly.
    """


class VDeviceOverwritingError(BalderException):
    """
    is thrown if the device is not overwritten correctly
    """


class AccessToUnmappedVDeviceException(BalderException):
    """
    is thrown if an unmapped vdevice is accessed
    """


class DeviceOverwritingError(BalderException):
    """
    is thrown if the device is not overwritten correctly
    """


class FeatureOverwritingError(BalderException):
    """
    is thrown if an inner instantiated feature is not overwritten correctly
    """


class UnknownVDeviceException(BalderException):
    """
    is thrown if some not allowed vDevices are given in a :meth:`Feature`
    """


class RoutingBrokenChainError(BalderException):
    """
    is thrown if the routing determines a broken pipe
    """


class IllegalConnectionTypeError(BalderException):
    """
    is thrown if the connection is illegal
    """


class ConnectionMetadataConflictError(BalderException):
    """
    is thrown if here is an illegal conflict between some metadata
    """


class DeviceScopeError(BalderException):
    """
    is thrown if the device class is placed on an illegal scope
    """


class ConnectionIntersectionError(BalderException):
    """
    is thrown if the intersection connection between two devices is reduced in a way that no intersection exists anymore
    """


class UnclearAssignableFeatureConnectionError(BalderException):
    """
    is thrown if two devices have parallel connections and a used feature could use both or more at the same time
    """


class InheritanceError(BalderException):
    """
    is thrown if a class inheritance is not allowed for the mentioned class
    """


class MultiInheritanceError(BalderException):
    """
    is thrown if multi inheritance was used where it is not allowed
    """


class InnerFeatureResolvingError(BalderException):
    """
    is thrown if an inner feature reference can not be resolved with the instantiated features of the device
    """


class VDeviceResolvingError(BalderException):
    """
    is thrown if there is an error while resolving VDevice's.
    """


class IllegalVDeviceMappingError(BalderException):
    """
    is thrown if there is an error while resolving vDevice mappings.
    """


class MissingFeaturesOfVDeviceError(BalderException):
    """
    is thrown if the related device does not implement all features specified in its mapped VDevice.
    """


class NotApplicableVariationException(BalderException):
    """
    is thrown internally after the current variation is not applicable
    """


class UnclearMethodVariationError(BalderException):
    """
    is thrown if there are more than one possible method variations possible
    """


class UnexpectedPluginMethodReturnValue(BalderException):
    """
    is thrown if a user plugin doesn't return something or returns wrong values in its plugin method
    """



import re

from _balder.testresult import ResultState
from _balder.balder_session import BalderSession
from ...test_utilities.base_0_envtester_class import Base0EnvtesterClass


class Test0ResolveOnly(Base0EnvtesterClass):
    """
    This testcase executes the basic ENV example with a specific command line argument ``--resolve-only``.

    The test checks that no tests are executed and the printed information is like expected.
    """

    @property
    def cmd_args(self):
        return ['--resolve-only']

    @property
    def expected_data(self) -> tuple:
        return tuple()

    def validate_printed_output(self, stdout: str) -> bool:
        assert self._check_header_of_stdout(stdout, 2, 2, 2), f"problems within header output"
        remaining_lines = stdout.splitlines()
        assert remaining_lines[7] == "RESOLVING OVERVIEW"
        split_blocks = "\n".join(remaining_lines[8:]).strip().split('\n\n')
        assert len(split_blocks) == 2

        expected_blocks = [
"""++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
+ [APPLICABLE] Scenario `ScenarioB` <-> Setup `SetupB`
+    ScenarioB.ScenarioDevice3 = SetupB.SetupDevice3
+    ScenarioB.ScenarioDevice4 = SetupB.SetupDevice4
+    -> Testcase<ScenarioB.test_b_1>
+    -> Testcase<ScenarioB.test_b_2>
++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++""",
"""++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
+ [APPLICABLE] Scenario `ScenarioA` <-> Setup `SetupA`
+    ScenarioA.ScenarioDevice1 = SetupA.SetupDevice1
+    ScenarioA.ScenarioDevice2 = SetupA.SetupDevice2
+    -> Testcase<ScenarioA.test_a_1>
+    -> Testcase<ScenarioA.test_a_2>
++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++"""
        ]

        assert sorted(split_blocks) == sorted(expected_blocks), f"detect issues in blocks"
        return True

    @staticmethod
    def validate_finished_session(session: BalderSession):
        # check result states everywhere (have to be SUCCESS everywhere
        assert session.executor_tree.executor_result == ResultState.NOT_RUN, \
            "test session does not terminates with NOT_RUN"



from multiprocessing import Process

from _balder.testresult import ResultState
from _balder.balder_session import BalderSession


def test_1_reverse_connect_3_scen_4_setup_devs(balder_working_dir):
    """
    This testcase uses the calculator environment in a modified version. It gets an additional `NumberProvider3` on
    setup level. On setup level all connections are defined over the `Calculator` (the connections are defined reverse).

    The test validates that the session terminates successfully and all three variations were executed.
    """
    proc = Process(target=processed, args=(balder_working_dir, ))
    proc.start()
    proc.join()
    assert proc.exitcode == 0, "the process terminates with an error"


def processed(env_dir):
    print("\n", flush=True)
    session = BalderSession(cmd_args=[], working_dir=env_dir)
    session.run()
    print("\n")

    assert session.executor_tree.executor_result == ResultState.SUCCESS, \
        "test session does not terminates with success"
    assert len(session.executor_tree.get_setup_executors()) == 1, \
        "found not exactly one setup executor"

    scenario_executors = session.executor_tree.get_setup_executors()[0].get_scenario_executors()

    assert len(scenario_executors) == 1, \
        "found not exactly one scenario executor"
    assert len(scenario_executors[0].get_variation_executors()) == 6, \
        "found not exactly three variation executor"



import balder
from .connections import SimulatedParentConnection, SimulatedChildConnection, MySimplySharedMemoryConnection


class ProvidesANumberFeature(balder.Feature):

    the_number = 0

    def set_number(self, number):
        """user method that allows to set the number"""
        self.the_number = number

    def get_number(self):
        """returns the set number"""
        raise NotImplementedError("has to be implemented in subclass")

    def sends_the_number(self):
        """sends the set number"""
        raise NotImplementedError("has to be implemented in subclass")


class AddCalculateFeature(balder.Feature):

    all_numbers = []

    def get_numbers(self):
        """this method get all the single numbers that should be calculated"""
        raise NotImplementedError("has to be implemented in subclass")

    def add_numbers(self):
        """this method adds all the numbers"""
        raise NotImplementedError("has to be implemented in subclass")


# the following connection is CONTAINED IN the scenario connection
@balder.for_vdevice('CalculatorDevice', SimulatedChildConnection.based_on(SimulatedParentConnection))
class VDeviceHelperFeature1(balder.Feature):
    """This is a helper feature that will be used to simulate this test."""
    class CalculatorDevice(balder.VDevice):
        add = AddCalculateFeature()


# the following connection is CONTAINED IN the scenario connection
@balder.for_vdevice('CalculatorDevice', SimulatedChildConnection)
class VDeviceHelperFeature2(balder.Feature):
    """This is a helper feature that will be used to simulate this test."""

    class CalculatorDevice(balder.VDevice):
        add = AddCalculateFeature()


# the following connection is CONTAINED IN the scenario connection
@balder.for_vdevice('CalculatorDevice', MySimplySharedMemoryConnection)
class VDeviceHelperFeature3(balder.Feature):
    """This is a helper feature that will be used to simulate this test."""

    class CalculatorDevice(balder.VDevice):
        add = AddCalculateFeature()



from multiprocessing import Process

from _balder.testresult import ResultState
from _balder.balder_session import BalderSession
from balder.exceptions import FeatureOverwritingError


def test_1_scenario_inheritance_wrong_feature_overwrite(balder_working_dir):
    """
    This testcase executes the calculator environment, but uses a special child scenario class that overwrites a feature
    property of an inherited device which uses a complete independent feature. This is not allowed and should result in
    a collector error.
    """
    proc = Process(target=processed, args=(balder_working_dir, ))
    proc.start()
    proc.join()
    assert proc.exitcode == 0, "the process terminates with an error"


def processed(env_dir):
    print("\n", flush=True)
    session = BalderSession(cmd_args=[], working_dir=env_dir)
    try:
        session.run()
        print("\n")
        assert False, "test session terminates without an error"
    except FeatureOverwritingError as exc:
        assert exc.args[0] == "the feature `AddCalculateFeature` with the attribute name `adds` of the device " \
                              "`Calculator` you are trying to overwrite is no child class of the feature " \
                              "`IllegalFeature` that was assigned to this property before"

    assert session.executor_tree is None, "test session does not terminates before collector work was done"



from __future__ import annotations
from typing import Type, Dict, Union, TYPE_CHECKING

import logging
from _balder.setup import Setup
from _balder.exceptions import IllegalVDeviceMappingError, MultiInheritanceError
from _balder.controllers.feature_controller import FeatureController
from _balder.controllers.device_controller import DeviceController
from _balder.controllers.normal_scenario_setup_controller import NormalScenarioSetupController

if TYPE_CHECKING:
    pass

logger = logging.getLogger(__file__)


class SetupController(NormalScenarioSetupController):
    """
    This is the controller class for :class:`Setup` items.
    """

    # helper property to disable manual constructor creation
    __priv_instantiate_key = object()

    #: contains all existing setups and its corresponding controller object
    _items: Dict[Type[Setup], SetupController] = {}

    def __init__(self, related_cls, _priv_instantiate_key):

        # describes if the current controller is for setups or for scenarios (has to be set in child controller)
        self._related_type = Setup

        # this helps to make this constructor only possible inside the controller object
        if _priv_instantiate_key != SetupController.__priv_instantiate_key:
            raise RuntimeError('it is not allowed to instantiate a controller manually -> use the static method '
                               '`SetupController.get_for()` for it')

        if not isinstance(related_cls, type):
            raise TypeError('the attribute `related_cls` has to be a type (no object)')
        if not issubclass(related_cls, Setup):
            raise TypeError(f'the attribute `related_cls` has to be a sub-type of `{Setup.__name__}`')
        if related_cls == Setup:
            raise TypeError(f'the attribute `related_cls` is `{Setup.__name__}` - controllers for native type are '
                            f'forbidden')
        # contains a reference to the related class this controller instance belongs to
        self._related_cls = related_cls

    # ---------------------------------- STATIC METHODS ----------------------------------------------------------------

    @staticmethod
    def get_for(related_cls: Type[Setup]) -> SetupController:
        """
        This class returns the current existing controller instance for the given item. If the instance does not exist
        yet, it will automatically create it and saves the instance in an internal dictionary.
        """
        if SetupController._items.get(related_cls) is None:
            item = SetupController(related_cls, _priv_instantiate_key=SetupController.__priv_instantiate_key)
            SetupController._items[related_cls] = item

        return SetupController._items.get(related_cls)

    # ---------------------------------- CLASS METHODS -----------------------------------------------------------------

    # ---------------------------------- PROPERTIES --------------------------------------------------------------------

    @property
    def related_cls(self) -> Type[Setup]:
        return self._related_cls

    # ---------------------------------- PROTECTED METHODS -------------------------------------------------------------

    # ---------------------------------- METHODS -----------------------------------------------------------------------

    def get_next_parent_class(self) -> Union[Type[Setup], None]:
        """
        This method returns the next parent class which is a subclass of the :class:`Setup` itself.

        :return: returns the next parent class or None if the next parent class is :class:`Setup`
                 itself
        """
        next_base_class = None
        for cur_base in self.related_cls.__bases__:
            if issubclass(cur_base, Setup):
                if next_base_class is not None:
                    raise MultiInheritanceError(
                        f"found more than one Setup parent classes for `{self.related_cls.__name__}` "
                        f"- multi inheritance is not allowed for Scenario/Setup classes")
                next_base_class = cur_base
        if next_base_class == Setup:
            return None
        return next_base_class

    def validate_feature_possibility(self):
        """
        This method validates that every feature connection (that already has a vDevice<->Device mapping on setup level)
        has a connection that is CONTAINED-IN the connection of the related setup devices.
        """
        all_devices = self.get_all_abs_inner_device_classes()
        for cur_device in all_devices:
            cur_device_instantiated_features = \
                DeviceController.get_for(cur_device).get_all_instantiated_feature_objects()
            for _, cur_feature in cur_device_instantiated_features.items():
                mapped_vdevice, mapped_device = cur_feature.active_vdevice_device_mapping
                if mapped_device is None:
                    # ignore this, because we have no vDevice mapping on setup level
                    continue

                cur_feature_controller = FeatureController.get_for(cur_feature.__class__)
                feature_class_based_for_vdevice = cur_feature_controller.get_abs_class_based_for_vdevice()
                if not feature_class_based_for_vdevice or mapped_vdevice not in feature_class_based_for_vdevice.keys():
                    # there exists no class based for vdevice information (at least for the current active vdevice)
                    continue

                # there exists a class based requirement for this vDevice
                class_based_cnn = feature_class_based_for_vdevice[mapped_vdevice]
                # search relevant connection
                cur_device_controller = DeviceController.get_for(cur_device)
                for _, cur_cnn_list in cur_device_controller.get_all_absolute_connections().items():
                    for cur_cnn in cur_cnn_list:
                        if not cur_cnn.has_connection_from_to(cur_device, end_device=mapped_device):
                            # this connection can be ignored, because it is no connection between the current device
                            # and the mapped device
                            continue
                        # check if the class-based feature connection is CONTAINED-IN this
                        # absolute-connection
                        if not class_based_cnn.contained_in(cur_cnn, ignore_metadata=True):

                            raise IllegalVDeviceMappingError(
                                f"the @for_vdevice connection for vDevice `{mapped_vdevice.__name__}` "
                                f"of feature `{cur_feature.__class__.__name__}` (used in "
                                f"`{cur_device.__qualname__}`) uses a connection that does not fit "
                                f"with the connection defined in setup class "
                                f"`{cur_device_controller.get_outer_class().__name__}` to related "
                                f"device `{mapped_device.__name__}`")



from __future__ import annotations

from _balder.controllers.device_controller import DeviceController
from _balder.device import Device
from _balder.node_gateway import NodeGateway


def gateway(from_node: str, to_node: str, bidirectional: bool = True):
    """
    This decorator enables two nodes of a device to be connected to one another via a gateway. The gateway can
    implement an unidirectional or bidirectional connection.
    """
    if not isinstance(from_node, str) and not isinstance(to_node, str):
        raise ValueError("the value of `from_node` and `to_node` must be the name of the nodes (type `str`)")

    if isinstance(bidirectional, bool):
        raise ValueError("the value of `bidirectional` must be a `bool`")

    def decorator(cls):
        nonlocal from_node
        nonlocal to_node
        nonlocal bidirectional

        # it must be a class decorator
        if not issubclass(cls, Device):
            raise TypeError(
                f"The decorator `gateway` may only be used for `Device` objects. This is not possible for the applied "
                f"class `{cls.__name__}`.")
        decorated_cls_device_controller = DeviceController.get_for(cls)

        new_gateway = NodeGateway(cls, from_node, to_node, bidirectional)
        decorated_cls_device_controller.add_new_raw_gateway(new_gateway)
        return cls
    return decorator



import balder
from ...lib.features import ScenarioMethVarFeature
from ...balderglob import RuntimeObserver


@balder.for_vdevice("VDeviceI", with_connections=balder.Connection())
@balder.for_vdevice("VDeviceII", with_connections=balder.Connection())
@balder.for_vdevice("VDeviceIII", with_connections=balder.Connection())
class BetweenMethVarFeature(ScenarioMethVarFeature):
    """This is the feature class that is between the scenario class and the setup class."""
    class VDeviceI(ScenarioMethVarFeature.VDeviceI):
        pass

    class VDeviceII(ScenarioMethVarFeature.VDeviceII):
        pass

    class VDeviceIII(ScenarioMethVarFeature.VDeviceIII):
        pass

    def do_something(self):
        RuntimeObserver.add_entry(
            __file__, BetweenMethVarFeature, BetweenMethVarFeature.do_something,
            "enter `BetweenMethVarFeature.do_something`",
            category="feature")

    @balder.for_vdevice("VDeviceI", with_connections=balder.Connection())
    def do_something_as_var(self):
        super().do_something_as_var()
        RuntimeObserver.add_entry(
            __file__, BetweenMethVarFeature, BetweenMethVarFeature.do_something_as_var,
            "MethodVariation<VDeviceI><BetweenMethVarFeature.do_something_as_var>", category="feature")

    @balder.for_vdevice("VDeviceII", with_connections=balder.Connection())
    def do_something_as_var(self):
        super().do_something_as_var()
        RuntimeObserver.add_entry(
            __file__, BetweenMethVarFeature, BetweenMethVarFeature.do_something_as_var,
            "MethodVariation<VDeviceII><BetweenMethVarFeature.do_something_as_var>", category="feature")

    @balder.for_vdevice("VDeviceIII", with_connections=balder.Connection())
    def do_something_as_var(self):
        super().do_something_as_var()
        RuntimeObserver.add_entry(
            __file__, BetweenMethVarFeature, BetweenMethVarFeature.do_something_as_var,
            "MethodVariation<VDeviceIII><BetweenMethVarFeature.do_something_as_var>", category="feature")




from __future__ import annotations
from typing import Iterable, Any

import inspect
from _balder.collector import Collector


def parametrize(
        field_name: str,
        values: Iterable[Any],
):
    """
    Allows to parametrize a test function. This decorator will be used to statically parametrize a test function.

    :param field_name: the field name of the test function

    :param values: an iterable of values, that should be used to parametrize the test function
    """
    if not isinstance(field_name, str):
        raise ValueError('the given field name must be a string')

    def decorator(func):
        nonlocal field_name
        nonlocal values

        if not inspect.isfunction(func):
            raise TypeError('the decorated object needs to be a test method')

        Collector.register_possible_parametrization(func, field_name, values)
        return func
    return decorator



from __future__ import annotations

import copy
from typing import Type, Dict, Union, List, Callable, Tuple

import logging
import inspect
from _balder.cnnrelations import OrConnectionRelation
from _balder.device import Device
from _balder.vdevice import VDevice
from _balder.feature import Feature
from _balder.controllers import Controller
from _balder.controllers.vdevice_controller import VDeviceController
from _balder.connection import Connection
from _balder.exceptions import UnclearMethodVariationError, MultiInheritanceError, VDeviceOverwritingError, \
    VDeviceResolvingError, FeatureOverwritingError, IllegalVDeviceMappingError

logger = logging.getLogger(__file__)


class FeatureController(Controller):
    """
    This is the controller class for :class:`Feature` items.
    """
    # helper property to disable manual constructor creation
    __priv_instantiate_key = object()

    #: contains all existing feature and its corresponding controller object
    _items: Dict[Type[Feature], FeatureController] = {}

    def __init__(self, related_cls, _priv_instantiate_key):

        # this helps to make this constructor only possible inside the controller object
        if _priv_instantiate_key != FeatureController.__priv_instantiate_key:
            raise RuntimeError('it is not allowed to instantiate a controller manually -> use the static method '
                               '`FeatureController.get_for()` for it')

        if not isinstance(related_cls, type):
            raise TypeError('the attribute `related_cls` has to be a type (no object)')
        if not issubclass(related_cls, Feature):
            raise TypeError(f'the attribute `related_cls` has to be a sub-type of `{Feature.__name__}`')
        if related_cls == Feature:
            raise TypeError(f'the attribute `related_cls` is `{Feature.__name__}` - controllers for native type are '
                            f'forbidden')
        # contains a reference to the related class this controller instance belongs to
        self._related_cls = related_cls

        #: holds the defined **Class-Based-Binding** for the related feature class sorted by VDevice types
        self._cls_for_vdevice: Dict[Type[VDevice], Connection] = {}

        #: holds the absolute calculated **Class-Based-Binding** for the related feature class sorted by VDevice types
        #: (will be calculated by :meth:`FeatureController.determine_absolute_class_based_for_vdevice`, which will be
        #: called during collecting)
        self._abs_cls_for_vdevice: Union[Dict[Type[VDevice], Connection], None] = None

        #: contains the **Method-Based-Binding** information for the current feature type (will be automatically set by
        #: executor)
        self._for_vdevice: Union[Dict[str, Dict[Callable, Dict[Type[VDevice], Connection]]], None] = None

        #: contains the original defined :class:`VDevice` objects for this feature (will be automatically set by
        #:  :class:`Collector`)
        self._original_vdevice_definitions: Union[Dict[str, Type[VDevice]], None] = None

        #: contains the current active method variations for the related feature class - for every key (method name str)
        #: a tuple with the VDevice type, the valid connection and the callable itself will be returned
        self._current_active_method_variation: Dict[str, Tuple[Type[VDevice], Connection, Callable]] = {}

    # ---------------------------------- STATIC METHODS ----------------------------------------------------------------

    @staticmethod
    def get_for(related_cls: Type[Feature]) -> FeatureController:
        """
        This class returns the current existing controller instance for the given item. If the instance does not exist
        yet, it will automatically create it and saves the instance in an internal dictionary.
        """
        if FeatureController._items.get(related_cls) is None:
            item = FeatureController(related_cls, _priv_instantiate_key=FeatureController.__priv_instantiate_key)
            FeatureController._items[related_cls] = item

        return FeatureController._items.get(related_cls)

    # ---------------------------------- CLASS METHODS -----------------------------------------------------------------

    # ---------------------------------- PROPERTIES --------------------------------------------------------------------

    @property
    def related_cls(self) -> Type[Feature]:
        return self._related_cls

    # ---------------------------------- PROTECTED METHODS -------------------------------------------------------------

    def _validate_vdevice_reference_used_in_for_vdevice_decorators(self):
        # now check if a definition for this class exists
        all_vdevices = self.get_abs_inner_vdevice_classes()
        # check the class based @for_vdevice and check the used vDevice classes here
        for cur_decorated_vdevice in self.get_class_based_for_vdevice().keys():
            if cur_decorated_vdevice not in all_vdevices:
                raise VDeviceResolvingError(
                    f"you assign a vDevice to the class based decorator `@for_vdevice()` of the feature class "
                    f"`{self.related_cls.__name__}` which is no direct member of this feature - note that you have "
                    f"to define the vDevice in your feature before using it in the decorator - if necessary "
                    f"overwrite it")
        # check the method based @for_vdevice and check the used vDevice classes here
        if self.get_method_based_for_vdevice() is not None:
            for cur_method_name, cur_method_data in self.get_method_based_for_vdevice().items():
                for _, vdevice_dict in cur_method_data.items():
                    for cur_vdevice in vdevice_dict.keys():
                        if cur_vdevice not in all_vdevices:
                            raise VDeviceResolvingError(
                                f"you assign a vDevice to the method variation `{cur_method_name}` of the feature "
                                f"class `{self.related_cls.__name__}` which is no direct member of this feature - note "
                                f"that you have to use the overwritten version if you overwrite a vDevice")

    def _get_method_based_for_vdevice_intersection(self, for_vdevice) -> List[Connection]:
        """helper method that determines the intersection connection list of all method based `@for_vdevice`
        connections of the given `for_vdevice`"""
        intersection = []

        if self.get_method_based_for_vdevice() is not None:
            for _, method_dict in self.get_method_based_for_vdevice().items():
                for _, vdevice_dict in method_dict.items():
                    if for_vdevice not in vdevice_dict.keys():
                        continue
                    for cur_cnn in vdevice_dict[for_vdevice]:
                        if isinstance(cur_cnn, type):
                            cur_cnn = cur_cnn()
                        # clean metadata here because this is no connection between real devices
                        cur_cnn.set_metadata_for_all_subitems(None)
                        intersection.append(cur_cnn)
        if len(intersection) == 0:
            return [Connection()]
        return intersection

    def _determine_all_theoretically_unordered_method_variations(
            self, of_method_name: str, for_vdevice: Type[VDevice],
            with_connection: Union[Connection, Tuple[Connection]]) -> Dict[Callable, Connection]:
        """
        This method returns all theoretically matching method variations. It returns more than one, if there are
        multiple method variation for the given VDevice in this feature, where the given connection is part of the
        connection described by the method variation.

        :param of_method_name: the name of the method that should be returned
        :param for_vdevice: the VDevice that is mapped
        :param with_connection: the connection that is used between the device that uses the related feature and the
                                VDevice
        :return: a dictionary that holds all available method variation that matches here
        """
        all_vdevice_method_variations = self.get_method_based_for_vdevice()

        if all_vdevice_method_variations is None:
            raise ValueError("the current feature has no method variations")

        if of_method_name not in all_vdevice_method_variations.keys():
            raise ValueError(f"can not find the method `{of_method_name}` in method variation data dictionary")

        all_possible_method_variations = {}
        for cur_impl_method, cur_method_impl_dict in all_vdevice_method_variations[of_method_name].items():
            if for_vdevice in cur_method_impl_dict.keys():
                cur_impl_method_cnns = cur_method_impl_dict[for_vdevice].get_singles()
                for cur_single_impl_method_cnn in cur_impl_method_cnns:
                    if cur_single_impl_method_cnn.contained_in(with_connection, ignore_metadata=True):
                        # this variation is possible
                        # ADD IT if it is not available yet
                        if cur_impl_method not in all_possible_method_variations.keys():
                            all_possible_method_variations[cur_impl_method] = cur_single_impl_method_cnn
                        # COMBINE IT if it is already available
                        else:
                            all_possible_method_variations[cur_impl_method] = Connection.based_on(
                                OrConnectionRelation(all_possible_method_variations[cur_impl_method],
                                                     cur_single_impl_method_cnn))
        return all_possible_method_variations

    # ---------------------------------- METHODS -----------------------------------------------------------------------

    def get_class_based_for_vdevice(self) -> Dict[Type[VDevice], Connection]:
        """
        This method returns the class based data for the `@for_vdevice` decorator.
        """
        return copy.copy(self._cls_for_vdevice)

    def get_abs_class_based_for_vdevice(self) -> Dict[Type[VDevice], Connection]:
        """
        This method returns the absolute calculated class-based-for-vdevice data for this feature.
        """
        if self._abs_cls_for_vdevice is None:
            raise RuntimeError('can not access the absolute class based for-vdevices because they are not set yet')
        return self._abs_cls_for_vdevice

    def set_class_based_for_vdevice(self, data: Dict[Type[VDevice], Connection]):
        """
        This method allows to set the data of the class based `@for_vdevice` decorator.
        """
        self._cls_for_vdevice = data

    def determine_absolute_class_based_for_vdevice(self, print_warning):
        """

        This method determines the absolute class based `@for_vdevice` value for the related feature.

        First it checks if there is a direct class based `@for_vdevice` decorator for this feature. It will not change
        anything, if the value was already set by an explicit class based `@for_vdevice` decorator. In this case the
        method only checks that every given vDevice class is a real part of the current :class:`Feature` class (will be
        returned by direct call of method `Feature.get_inner_vdevice_classes()`). Otherwise, it determines the class
        based `@for_vdevice` value through analysing of the method based decorators and sets this determined value. If
        the method has to determine the value, it throws a warning with a suggestion for a nice class based decorator.
        Also, here the method will analyse the given vDevice classes and secures that they are defined in the current
        :class:`Feature` class.

        .. note::
            This method automatically updates the values for the parent classes, too. Every time it searches for the
            values it considers the parent values for the vDevice or the parent class of the vDevice, too.

        .. note::
            This method can throw a user warning (`throw_warning` has to be True for that), but only on the given list
            of :class:`Feature` classes. All parent :class:`Feature` classes will be determined correctly, but will not
            throw a waring.
        """
        # first determine this for all parent classes
        next_parent_feat = self.get_next_parent_feature()
        # with the following recursive call we guarantee that the next parent class has the correct resolved class
        #  based @for_vdevice information
        if next_parent_feat:
            FeatureController.get_for(next_parent_feat).determine_absolute_class_based_for_vdevice(print_warning=False)

        # validate if all used vDevice references in method and class based `@for_vdevice` decorators can be used,
        # because they are members of this feature
        self._validate_vdevice_reference_used_in_for_vdevice_decorators()

        # now check if a definition for this class exists
        all_vdevices = self.get_abs_inner_vdevice_classes()

        cls_based_for_vdevice = self.get_class_based_for_vdevice()
        for cur_vdevice in all_vdevices:
            # determine the class based for_vdevice value only if there is no one defined for this vDevice
            if cur_vdevice in cls_based_for_vdevice.keys():
                # there already exists a definition for this vDevice -> IGNORE
                continue

            # first determine the valid parent intersection (can also be extended itself)

            # get the correct vDevice that is known in the parent feature class
            vdevice_controller = VDeviceController.get_for(cur_vdevice)
            if vdevice_controller.get_outer_class() == self.related_cls:
                # this cur_vdevice is defined in this Feature class -> check if a parent class of it exists in
                # parent classes of this feature
                #  -> search parent vDevice existence and check if this vDevice is also used in a parent class of
                #     this feature
                #  -> if it does not exist -> there is no parent class based definition
                vdevice_of_interest = vdevice_controller.get_next_parent_vdevice()

            else:
                # the definition scope of the VDevice is in a higher Feature parent -> has to be in the
                # `__cls_for_vdevice` in one of the next base classes
                vdevice_of_interest = cur_vdevice

            parent_values = []
            # read vDevice connection tree of parent feature class (only if the vDevice already exists in the
            # parent class)
            if vdevice_of_interest is not None and next_parent_feat is not None:
                next_parent_feat_cls_based_for_vdevice = \
                    FeatureController.get_for(next_parent_feat).get_abs_class_based_for_vdevice()
                if vdevice_of_interest in next_parent_feat_cls_based_for_vdevice.keys():
                    cnn = next_parent_feat_cls_based_for_vdevice[vdevice_of_interest]
                    # clean metadata here because this is no connection between real devices
                    cnn.set_metadata_for_all_subitems(None)
                    parent_values.append(cnn)

            this_vdevice_intersection = parent_values

            # determine the class value automatically by discovering all method variations for this vDevice only
            this_vdevice_intersection += self._get_method_based_for_vdevice_intersection(for_vdevice=cur_vdevice)

            # print warning only if the printing is enabled and there is a sub connection tree (otherwise, the
            #  decorator is not necessary)
            if print_warning and len(this_vdevice_intersection) > 0:
                # only print the warning if there exists method variations for the feature
                if self.get_method_based_for_vdevice() is not None:
                    logger.warning(
                        f"your used feature class `{self.related_cls.__name__}` doesn't provide a class based "
                        f"@for_vdevice decorator for the vDevice `{cur_vdevice.__name__}`\n"
                        f"Balder has determined a possible marker:\n\n"
                        f'@balder.for_vdevice("{cur_vdevice.__name__}", '
                        f'{", ".join([cur_cnn.get_tree_str() for cur_cnn in this_vdevice_intersection])})\n\n')

            # set the determined data into the class based `@for_vdevice` class property
            cls_based_for_vdevice[cur_vdevice] = Connection.based_on(OrConnectionRelation(*this_vdevice_intersection))

        self._abs_cls_for_vdevice = cls_based_for_vdevice

    def get_method_based_for_vdevice(self) -> \
            Union[Dict[str, Dict[Callable, Dict[Type[VDevice], Connection]]], None]:
        """
        This method returns the method based data for the `@for_vdevice` decorator or None, if there is no decorator
        given
        """
        return self._for_vdevice

    def set_method_based_for_vdevice(
            self, data: Union[Dict[str, Dict[Callable, Dict[Type[VDevice], Connection]]], None]):
        """
        This method allows to set the data for the method based `@for_vdevice` decorator.
        """
        self._for_vdevice = data

    def get_method_variation(
            self,
            of_method_name: str,
            for_vdevice: Type[VDevice],
            with_connection: Connection,
            ignore_no_findings: bool = False
    ) -> Union[Callable, None]:
        """
        This method searches for the unique possible method variation and returns it. In its search, the method also
        includes the parent classes of the related feature element of this controller.

        .. note::
            The method throws an exception if it can not find a valid unique method variation for the given data.

        .. note::
            Note, that the method does not check if the method name, the VDevice nor the given `connection` is really a
            part of this object. Please secure that the data is validated before.

        .. note::
            The method determines all possible method-variations. If it finds more than one clear method variation it
            tries to sort them hierarchical. This is done by checking if one possible method variation is contained in
            the other. If this can be clearly done, the method returns the furthest out one. Otherwise, it throws an
            `UnclearMethodVariationError`

        :param of_method_name: the name of the method that should be returned

        :param for_vdevice: the VDevice that is mapped

        :param with_connection: the connection that is used between the device that uses the related feature and the
                                VDevice

        :param ignore_no_findings: if this attribute is true, the method will not throw an exception if it can not find
                                   something, it only returns None

        :return: the method variation callable for the given data (or none, if the method does not exist in this object
                 or in a parent class of it)
        """
        # first determine all possible method-variations
        all_possible_method_variations = self._determine_all_theoretically_unordered_method_variations(
            of_method_name=of_method_name, for_vdevice=for_vdevice, with_connection=with_connection)

        # there are no method variations in this feature directly -> check parent classes
        if len(all_possible_method_variations) == 0:
            # try to execute this method in parent classes
            for cur_base in self.related_cls.__bases__:
                if issubclass(cur_base, Feature) and cur_base != Feature:
                    parent_meth_result = FeatureController.get_for(cur_base).get_method_variation(
                        of_method_name=of_method_name, for_vdevice=for_vdevice, with_connection=with_connection,
                        ignore_no_findings=True)
                    if parent_meth_result:
                        return parent_meth_result
            if not ignore_no_findings:
                raise UnclearMethodVariationError(
                    f"found no possible method variation for method "
                    f"`{self.related_cls.__name__}.{of_method_name}` with vDevice `{for_vdevice.__name__}` "
                    f"and usable connection `{with_connection.get_tree_str()}´")
            return None

        # we only have one -> selection is clear
        if len(all_possible_method_variations) == 1:
            return list(all_possible_method_variations.keys())[0]

        # if there are more than one possible method variation, try to determine the outer one
        remaining_possible_method_variations = list(
            filter(
                lambda meth: not max(all_possible_method_variations[meth].contained_in(cur_other_cnn)
                                     for cur_other_cnn in all_possible_method_variations.values()
                                     if cur_other_cnn != all_possible_method_variations[meth]),
                all_possible_method_variations.keys())
        )

        if len(remaining_possible_method_variations) > 1:
            raise UnclearMethodVariationError(
                f"found more than one possible method variation for method "
                f"`{self.related_cls.__name__}.{of_method_name}` with vDevice `{for_vdevice.__name__}` "
                f"and usable connection `{with_connection.get_tree_str()}´")
        return remaining_possible_method_variations[0]

    def get_inner_vdevice_classes(self) -> List[Type[VDevice]]:
        """
        This is a method that determines the inner VDevice classes for the related feature class. If the method can not
        find some VDevices in the current class it returns an empty list. This method will never search in parent
        classes.

        If you want to get the absolute VDevices use :meth:`Feature.get_inner_vdevice_classes`.
        """

        all_classes = inspect.getmembers(self.related_cls, inspect.isclass)
        filtered_classes = []
        for _, cur_class in all_classes:
            if not issubclass(cur_class, VDevice):
                # filter all classes and make sure that only the child classes of :class:`VDevice` remain
                continue

            if VDeviceController.get_for(cur_class).get_outer_class() != self.related_cls:
                # filter all classes that do not match the setup name in __qualname__
                continue
            # otherwise, add this candidate
            filtered_classes.append(cur_class)

        if len(filtered_classes) == 0:
            # do not found some VDevice classes -> search in parent
            for cur_base in self.related_cls.__bases__:
                if cur_base == Feature:
                    return []
                if issubclass(cur_base, Feature):
                    return FeatureController.get_for(cur_base).get_abs_inner_vdevice_classes()

        return filtered_classes

    def get_inner_vdevice_class_by_string(self, device_str: str) -> Union[Type[VDevice], None]:
        """
        This method returns the inner VDevice class for the given string.

        :param device_str: the name string of the VDevice that should be returned

        :return: the VDevice class or None, if the method has not found any class with this name
        """
        possible_vdevs = [cur_vdevice for cur_vdevice in self.get_inner_vdevice_classes()
                          if cur_vdevice.__name__ == device_str]
        if len(possible_vdevs) == 0:
            return None
        if len(possible_vdevs) > 1:
            raise RuntimeError("found more than one possible vDevices - something unexpected happened")

        return possible_vdevs[0]

    def get_abs_inner_vdevice_classes(self) -> List[Type[VDevice]]:
        """
        This is a method that determines the inner VDevice classes for the feature class. If the method can not find
        some VDevices in the related feature class it also starts searching in the base classes. It always returns the
        first existing definition in the relevant parent classes.
        """

        filtered_classes = self.get_inner_vdevice_classes()

        if len(filtered_classes) == 0:
            # do not found some VDevice classes -> search in parent
            for cur_base in self.related_cls.__bases__:
                if cur_base == Feature:
                    return []
                if issubclass(cur_base, Feature):
                    return FeatureController.get_for(cur_base).get_abs_inner_vdevice_classes()

        return filtered_classes

    def get_inner_referenced_features(self) -> Dict[str, Feature]:
        """
        This method returns a dictionary with all referenced :class:`Feature` objects, where the variable name is the
        key and the instantiated object the value.
        """

        result = {}
        for cur_name in dir(self.related_cls):
            cur_val = getattr(self.related_cls, cur_name)
            if isinstance(cur_val, Feature):
                result[cur_name] = cur_val
        return result

    def validate_inner_vdevice_inheritance(self):
        """
        This method validates the inheritance of all inner :class:`VDevice` classes of the feature that belongs to this
        controller.

        It secures that new :class:`VDevice` classes are added or existing :class:`VDevice` classes are completely being
        overwritten for every feature level. The method only allows the overwriting of :class:`VDevices`, which are
        subclasses of another :class:`VDevice` that is defined in a parent :class:`Feature` class. In addition, the
        class has to have the same name as its parent class.

        The method also secures that the user overwrites instantiated :class:`Feature` classes in the VDevice (class
        property name is the same) only with subclasses of the element that is being overwritten. New Features can be
        added without consequences.
        """

        all_direct_vdevices_of_this_feature_lvl = self.get_inner_vdevice_classes()
        if len(all_direct_vdevices_of_this_feature_lvl) != 0:
            # check that all absolute items of higher class are implemented
            next_feature_parent = None
            for cur_parent in self._related_cls.__bases__:
                if issubclass(cur_parent, Feature) and cur_parent != Feature:
                    if next_feature_parent is not None:
                        raise MultiInheritanceError(
                            "can not select the next parent class, found more than one parent classes for feature "
                            f"`{self._related_cls.__name__}` that is a subclass of `{Feature.__name__}`")
                    next_feature_parent = cur_parent
            # only continue if the current feature has a parent class
            if next_feature_parent:
                # first check the parent feature (secure that the inheritance chain is valid first)
                parent_collector = FeatureController.get_for(next_feature_parent)
                parent_collector.validate_inner_vdevice_inheritance()

                # now continue with checking the inheritance between this feature and its direct parent
                parent_vdevices = parent_collector.get_abs_inner_vdevice_classes()
                # now check that every parent vDevice also exists in the current selection
                for cur_parent_vdevice in parent_vdevices:
                    direct_namings = [cur_item.__name__ for cur_item in all_direct_vdevices_of_this_feature_lvl]
                    # check that the parent vDevice exists in the direct namings
                    if cur_parent_vdevice.__name__ not in direct_namings:
                        raise VDeviceOverwritingError(
                            f"missing overwriting of parent VDevice class `{cur_parent_vdevice.__qualname__}` in "
                            f"feature class `{self._related_cls.__name__}` - if you overwrite one or more VDevice(s) "
                            f"you have to overwrite all!")

                    # otherwise check if inheritance AND feature overwriting is correct
                    cur_child_idx = direct_namings.index(cur_parent_vdevice.__name__)
                    related_child_vdevice = all_direct_vdevices_of_this_feature_lvl[cur_child_idx]
                    if not issubclass(related_child_vdevice, cur_parent_vdevice):
                        # inherit from a parent device, but it has not the same naming -> NOT ALLOWED
                        raise VDeviceOverwritingError(
                            f"the inner vDevice class `{related_child_vdevice.__qualname__}` has the same "
                            f"name than the vDevice `{cur_parent_vdevice.__qualname__}` - it should also "
                            f"inherit from it")
                    # todo check that feature overwriting inside the VDevice is correct
                    # now check that the vDevice overwrites the existing properties only in a proper manner (to
                    #  overwrite it, it has to have the same property name as the property in the next parent
                    #  class)
                    cur_vdevice_features = \
                        VDeviceController.get_for(related_child_vdevice).get_all_instantiated_feature_objects()
                    cur_vdevice_base_features = \
                        VDeviceController.get_for(cur_parent_vdevice).get_all_instantiated_feature_objects()
                    for cur_base_property_name, cur_base_feature_instance in cur_vdevice_base_features.items():
                        # now check that every base property is available in the current vDevice too - check
                        #  that the instantiated feature is the same or the feature of the child vDevice is a
                        #  child of it -> ignore it, if the child vDevice has more features than the base -
                        #   that doesn't matter
                        if cur_base_property_name not in cur_vdevice_features.keys():
                            raise VDeviceResolvingError(
                                f"can not find the property `{cur_base_property_name}` of "
                                f"parent vDevice `{cur_parent_vdevice.__qualname__}` in the "
                                f"current vDevice class `{related_child_vdevice.__qualname__}`")
                        cur_feature_instance = cur_vdevice_features[cur_base_property_name]
                        if not isinstance(cur_feature_instance, cur_base_feature_instance.__class__):
                            raise FeatureOverwritingError(
                                f"you are trying to overwrite an existing vDevice Feature property "
                                f"`{cur_base_property_name}` in vDevice `{related_child_vdevice.__qualname__}` "
                                f"from the parent vDevice class `{cur_parent_vdevice.__qualname__}` - this is "
                                f"only possible with a child (or with the same) feature class the parent "
                                f"uses (in this case the `{cur_base_feature_instance.__class__.__name__}`)")

    def validate_inherited_class_based_vdevice_cnn_subset(self):
        """
        This method checks that the class based for_vdevice values of a child :class:`Feature` class are contained_in
        the related VDevice defined in a parent :class:`Feature` class.
        """

        to_checking_parent_features = []

        feature_vdevices = self.get_abs_inner_vdevice_classes()
        for cur_vdevice in feature_vdevices:
            cur_vdevice_cls_cnn = self.get_abs_class_based_for_vdevice().get(cur_vdevice)
            # get parent class of vdevice
            relevant_parent_class = VDeviceController.get_for(cur_vdevice).get_next_parent_vdevice()

            # continue with next vdevice if no relevant parent class was found
            if relevant_parent_class is None:
                continue

            relevant_parent_class_controller = VDeviceController.get_for(relevant_parent_class)
            # only if there is a higher class which has to be considered
            parent_vdevice_feature = relevant_parent_class_controller.get_outer_class()
            if parent_vdevice_feature not in to_checking_parent_features:
                to_checking_parent_features.append(parent_vdevice_feature)
            parent_vdevice_cnn = \
                FeatureController.get_for(
                    parent_vdevice_feature).get_abs_class_based_for_vdevice()[relevant_parent_class]
            # check if VDevice connection elements are all contained in the parent connection
            if not cur_vdevice_cls_cnn.contained_in(parent_vdevice_cnn, ignore_metadata=True):
                raise VDeviceResolvingError(
                    f"the VDevice `{cur_vdevice.__name__}` is a child of the VDevice "
                    f"`{relevant_parent_class.__name__}`, which doesn't implements the connection of "
                    f"the child - the connection element `{cur_vdevice_cls_cnn.get_tree_str()})´ is not "
                    f"contained in the connection-tree of the parent VDevice")

        # validate inheritance levels for all features with parent VDevices as inner-classes
        for cur_feature in to_checking_parent_features:
            FeatureController.get_for(cur_feature).validate_inherited_class_based_vdevice_cnn_subset()

    def get_next_parent_feature(self) -> Union[Type[Feature], None]:
        """
        This method returns the next parent class of this feature, which is still a subclass of :class:`Feature`. If
        the next parent class is :class:`Feature`, None will be returned.

        :return: the parent Feature class or None if no parent exists
        """
        possible_parent_classes = []
        for cur_parent in self.related_cls.__bases__:
            if issubclass(cur_parent, Feature) and cur_parent != Feature:
                possible_parent_classes.append(cur_parent)

        if len(possible_parent_classes) > 1:
            raise MultiInheritanceError(
                f"the feature `{self.related_cls.__name__}` has more than one parent classes from type "
                f"`Feature` - this is not allowed")

        if len(possible_parent_classes) == 1:
            # we have found one parent feature class
            return possible_parent_classes[0]

        # we have no parent class
        return None

    def set_active_method_variation(self, method_selection: Dict[str, Tuple[Type[VDevice], Connection, Callable]]):
        """
        This method sets the active method variation selection for the related feature class.
        :param method_selection: the method selection that should be set
        """
        self._current_active_method_variation = method_selection

    def get_active_method_variation(self, method_name: str) \
            -> Union[Tuple[Type[VDevice], Connection, Callable], Tuple[None, None, None]]:
        """
        This method returns the current active method variation for the given `method_name` for the related fixture.

        .. note::
            Please note, this method only returns the set active method variation for this related feature only. It does
            not check parent classes of this feature.

        :param method_name: the name of the method the current active method variation should be returned

        :return: a tuple with the current active method selection or a tuple with `None` if no active method variation
                 exists on this feature class level
        """
        return self._current_active_method_variation.get(method_name, tuple([None, None, None]))

    def get_inherited_method_variation(self, parent_class: Type[Feature], method_var_name: str):
        """
        This method will determine the correct inherited method-variation for the current object. For this, it searches
        in the base classes of the given `parent_class` (which has to be a parent class of `self`) for the
        method-variation that should be called.
        It automatically detects if the parent class has a method-variation or is a single normal method. In case that
        the method is a single normal method, it will directly return it, otherwise it searches the correct
        method-variation according to the vDevice mapping of the current object and return the current active
        method-variation.

        :param parent_class: the parent class of this object, the method should start searching for the
                             `method_var_name` method (it searches in this class and all parents)

        :param method_var_name: the name of the method or of the method variation that should be returned
        """

        parent_class_controller = FeatureController.get_for(parent_class)
        if parent_class_controller.get_method_based_for_vdevice() is not None and \
                method_var_name in parent_class_controller.get_method_based_for_vdevice().keys():
            # the parent class has a method-variation -> get the current active version of it

            # first get the active data for the instantiated feature object
            active_vdevice, active_cnn_intersection, _ = self.get_active_method_variation(method_var_name)
            # get the vDevice object that is used in the given parent class
            if hasattr(parent_class, active_vdevice.__name__):
                parent_vdevice = getattr(parent_class, active_vdevice.__name__)
            else:
                return None

            # then determine the correct method variation according to the data of the instantiated object
            cur_method_variation = parent_class_controller.get_method_variation(
                of_method_name=method_var_name, for_vdevice=parent_vdevice,
                with_connection=active_cnn_intersection, ignore_no_findings=True)
            return cur_method_variation

        if hasattr(parent_class, method_var_name):
            # we found one normal method in this object
            return getattr(parent_class, method_var_name)

        # execute this method for all based and check if there is exactly one
        next_base_feature_class = parent_class_controller.get_next_parent_feature()
        if next_base_feature_class is None:
            return None

        return self.get_inherited_method_variation(next_base_feature_class, method_var_name)

    def validate_inner_classes(self):
        """
        This method validates all inner classes of the related feature and secures that none of these subclasses are
        subclass of :class:`Device` but not subclasses from :class:`VDevice`. Of course other inner-classes that are not
        required for balder are allowed too.
        """
        all_inner_classes = inspect.getmembers(self.related_cls, inspect.isclass)
        for cur_inner_name, cur_inner_class in all_inner_classes:
            if not issubclass(cur_inner_class, Device):
                # ignore this element
                continue
            # do only check the inner classes that inherits from `Device`
            if not issubclass(cur_inner_class, VDevice):
                raise VDeviceResolvingError(
                    f"the inner class `{cur_inner_class.__name__}` with name `{cur_inner_name}` is a child "
                    f"class of `Device` but not from `VDevice` as expected")
            cur_inner_class_instantiated_features = \
                VDeviceController.get_for(cur_inner_class).get_all_instantiated_feature_objects()
            for _, cur_vdevice_feature in cur_inner_class_instantiated_features.items():
                if cur_vdevice_feature.active_vdevices != {}:
                    raise IllegalVDeviceMappingError(
                        f"the feature `{cur_vdevice_feature.__class__.__name__}` you have instantiated in your "
                        f"vDevice `{cur_inner_class.__name__}` of feature `{self.related_cls.__name__}` "
                        f"has a own vDevice mapping - vDevice mappings are allowed for features on Devices "
                        f"only")

    def get_original_vdevice_definitions(self) -> Dict[str, Type[VDevice]]:
        """
        This method returns the :class:`VDevice` definitions that are the original definitions for this feature.
        """
        if self._original_vdevice_definitions is None:
            raise RuntimeError('can not access the original VDevice definitions before they were set with '
                               '`save_all_current_vdevice_references_as_originals`')
        return self._original_vdevice_definitions

    def save_all_current_vdevice_references_as_originals(self):
        """
        This method saves the current existing :class:`VDevice` definitions inside this feature as originals.
        """
        new_originals = self.get_abs_inner_vdevice_classes()
        self._original_vdevice_definitions = {cur_vdevice.__name__: cur_vdevice for cur_vdevice in new_originals}
